import copy
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from gymnasium import spaces

from agilerl.modules.base import EvolvableModule, ModuleDict, MutationType, mutation
from agilerl.modules.cnn import EvolvableCNN
from agilerl.modules.mlp import EvolvableMLP
from agilerl.typing import ArrayOrTensor
from agilerl.utils.evolvable_networks import (
    get_activation,
    is_image_space,
    tuple_to_dict_space,
)

ModuleType = Union[EvolvableModule, nn.Module]
SupportedEvolvableTypes = Union[EvolvableCNN, EvolvableMLP]
TupleOrDictSpace = Union[spaces.Tuple, spaces.Dict]
TupleOrDictObservation = Union[Dict[str, ArrayOrTensor], Tuple[ArrayOrTensor]]


class EvolvableMultiInput(EvolvableModule):
    """Evolvable multi-input network composed of an evolvable feature extractor for each observation key,
    and an optional MLP for the concatenated vector inputs. The extracted features are concatenated and
    passed through a final MLP to produce the output tensor.

    .. note::
        The user can inherit from this class and override the `forward` method to define a custom
        forward pass for the network, manipulating the observation dictionary as needed, and adding any
        non-evolvable additional layers needed for their specific problem.

    :param observation_space: Dictionary or Tuple space of observations.
    :type observation_space: spaces.Dict or spaces.Tuple
    :param num_outputs: Dimension of the output tensor.
    :type num_outputs: int
    :param channel_size: List of channel sizes for the convolutional layers.
    :type channel_size: List[int]
    :param kernel_size: List of kernel sizes for the convolutional layers.
    :type kernel_size: List[int]
    :param stride_size: List of stride sizes for the convolutional layers.
    :type stride_size: List[int]
    :param latent_dim: Dimension of the latent space representation. Default is 16.
    :type latent_dim: int, optional
    :param cnn_block_type: Type of convolutional block to use. Default is "Conv2d".
    :type cnn_block_type: Literal["Conv2d", "Conv3d"], optional
    :param sample_input: Sample input tensor for the CNN. Default is None.
    :type sample_input: Optional[dict[str, torch.Tensor]], optional
    :param vector_space_mlp: Whether to use an MLP for the vector spaces. This is done by concatenating the
        flattened observations and passing them through an `EvolvableMLP`. Default is False, whereby the
        observations are concatenated directly to the feature encodings before the final MLP.
    :type vector_space_mlp: bool, optional
    :param hidden_size: List of hidden sizes for the MLP. Default is None.
    :type hidden_size: List[int], optional
    :param init_dicts: Dictionary of initialization dictionaries for the feature extractors. Default is {}.
    :type init_dicts: Dict[str, Dict[str, Any]], optional
    :param output_activation: Activation function for the output layer. Default is None.
    :type output_activation: Optional[str], optional
    :param activation: Activation function for the module layers. Default is "ReLU".
    :type activation: str, optional
    :param min_hidden_layers: Minimum number of hidden layers for the MLP. Default is 1.
    :type min_hidden_layers: int, optional
    :param max_hidden_layers: Maximum number of hidden layers for the MLP. Default is 3.
    :type max_hidden_layers: int, optional
    :param min_mlp_nodes: Minimum number of nodes for the MLP. Default is 64.
    :type min_mlp_nodes: int, optional
    :param max_mlp_nodes: Maximum number of nodes for the MLP. Default is 1024.
    :type max_mlp_nodes: int, optional
    :param min_cnn_hidden_layers: Minimum number of hidden layers for the CNN. Default is 1.
    :type min_cnn_hidden_layers: int, optional
    :param max_cnn_hidden_layers: Maximum number of hidden layers for the CNN. Default is 6.
    :type max_cnn_hidden_layers: int, optional
    :param min_channel_size: Minimum channel size for the CNN. Default is 32.
    :type min_channel_size: int, optional
    :param max_channel_size: Maximum channel size for the CNN. Default is 256.
    :type max_channel_size: int, optional
    :param layer_norm: Whether to use layer normalization. Default is False.
    :type layer_norm: bool, optional
    :param noise_std: Standard deviation of the noise. Default is 0.5.
    :type noise_std: float, optional
    :param noisy: Whether to use noisy layers. Default is False.
    :type noisy: bool, optional
    :param init_layers: Whether to initialize the layers. Default is True.
    :type init_layers: bool, optional
    :param device: Device to use for the network. Default is "cpu".
    :type device: str, optional
    """

    feature_net: ModuleDict

    # Supported underlying spaces in passed Dict or Tuple space
    _SupportedSpaces = (spaces.Box, spaces.Discrete, spaces.MultiDiscrete)

    def __init__(
        self,
        observation_space: TupleOrDictSpace,
        num_outputs: int,
        channel_size: List[int],
        kernel_size: List[int],
        stride_size: List[int],
        latent_dim: int = 16,
        cnn_block_type: Literal["Conv2d", "Conv3d"] = "Conv2d",
        sample_input: Optional[TupleOrDictObservation] = None,
        vector_space_mlp: bool = False,
        hidden_size: Optional[List[int]] = None,
        init_dicts: Optional[Dict[str, Dict[str, Any]]] = None,
        output_activation: Optional[str] = None,
        activation: str = "ReLU",
        min_hidden_layers: int = 1,
        max_hidden_layers: int = 3,
        min_mlp_nodes: int = 64,
        max_mlp_nodes: int = 1024,
        min_cnn_hidden_layers: int = 1,
        max_cnn_hidden_layers: int = 6,
        min_channel_size: int = 32,
        max_channel_size: int = 256,
        min_latent_dim: int = 8,
        max_latent_dim: int = 128,
        layer_norm: bool = False,
        noisy: bool = False,
        noise_std: float = 0.5,
        init_layers: bool = True,
        device: str = "cpu",
        name: str = "multi_input",
    ):
        super().__init__(device)

        assert num_outputs > 0, "Number of outputs must be greater than 0."
        assert isinstance(
            observation_space, (spaces.Dict, spaces.Tuple)
        ), "Observation space must be a Dict or Tuple space."
        assert (
            latent_dim <= max_latent_dim
        ), "Latent dimension must be less than or equal to max latent dimension."
        assert (
            latent_dim >= min_latent_dim
        ), "Latent dimension must be greater than or equal to min latent dimension."

        subspaces = (
            observation_space.spaces.values()
            if isinstance(observation_space, spaces.Dict)
            else observation_space.spaces
        )
        assert all(
            [isinstance(space, self._SupportedSpaces) for space in subspaces]
        ), "Observation space must contain only Box, Discrete, or MultiDiscrete spaces."

        if cnn_block_type == "Conv3d":
            if isinstance(observation_space, spaces.Tuple) and not isinstance(
                sample_input, tuple
            ):
                raise TypeError(
                    "Sample input must be provided as a tuple for Conv3d block type."
                )

            if isinstance(observation_space, spaces.Dict) and not isinstance(
                sample_input, dict
            ):
                raise TypeError(
                    "Sample input must be provided as a dict for Conv3d block type."
                )

        if vector_space_mlp:
            assert (
                hidden_size is not None
            ), "Hidden size must be specified for vector space MLP."

        # Convert Tuple space to Dict space for consistency
        self.is_tuple_space = False
        if isinstance(observation_space, spaces.Tuple):
            observation_space = tuple_to_dict_space(observation_space)
            self.is_tuple_space = True

        self._init_dicts = init_dicts if init_dicts is not None else {}
        self._activation = activation
        self.observation_space = observation_space
        self.num_outputs = num_outputs
        self.name = name
        self.vector_space_mlp = vector_space_mlp
        self.channel_size = channel_size
        self.kernel_size = kernel_size
        self.stride_size = stride_size
        self.hidden_size = hidden_size
        self.sample_input = sample_input
        self.cnn_block_type = cnn_block_type
        self.output_activation = output_activation
        self.min_hidden_layers = min_hidden_layers
        self.max_hidden_layers = max_hidden_layers
        self.min_mlp_nodes = min_mlp_nodes
        self.max_mlp_nodes = max_mlp_nodes
        self.min_latent_dim = min_latent_dim
        self.max_latent_dim = max_latent_dim
        self.min_cnn_hidden_layers = min_cnn_hidden_layers
        self.max_cnn_hidden_layers = max_cnn_hidden_layers
        self.min_channel_size = min_channel_size
        self.max_channel_size = max_channel_size
        self.layer_norm = layer_norm
        self.init_layers = init_layers
        self.latent_dim = latent_dim
        self.noisy = noisy
        self.noise_std = noise_std
        self.vector_spaces = [
            key
            for key, space in observation_space.spaces.items()
            if not is_image_space(space)
        ]

        self.feature_net = self.build_feature_extractor()

        # Collect all vector space shapes for concatenation
        vector_input_dim = sum(
            [
                spaces.flatdim(self.observation_space.spaces[key])
                for key in self.vector_spaces
            ]
        )

        # Calculate total feature dimension for final MLP
        self.image_features_dim = sum(
            [
                self.latent_dim
                for subspace in self.observation_space.spaces.values()
                if is_image_space(subspace)
            ]
        )

        self.vector_features_dim = (
            self.latent_dim if self.vector_space_mlp else vector_input_dim
        )
        features_dim = self.image_features_dim + self.vector_features_dim

        # Final dense layer to convert feature encodings to desired num_outputs
        self.final_dense = nn.Linear(features_dim, num_outputs, device=device)
        self.output = get_activation(output_activation)

    @property
    def net_config(self) -> Dict[str, Any]:
        """Returns the configuration of the network.

        :return: Network configuration
        :rtype: Dict[str, Any]
        """
        net_config = self.init_dict.copy()
        for attr in ["observation_space", "num_outputs", "device", "name"]:
            net_config.pop(attr)
        return net_config

    @property
    def activation(self) -> str:
        """Returns the activation function for the network.

        :return: Activation function
        :rtype: str
        """
        return self._activation

    @activation.setter
    def activation(self, activation: str) -> None:
        """Set the activation function for the network.

        :param activation: Activation function to use.
        :type activation: str
        """
        self._activation = activation
        for module in self.feature_net.values():
            module.change_activation(activation, output=True)

    @property
    def base_init_dict(self) -> Dict[str, Any]:
        """Returns dictionary of base information.

        :return: Base information
        :rtype: Dict[str, Any]
        """
        return {
            "num_outputs": self.latent_dim,
            "layer_norm": self.layer_norm,
            "init_layers": self.init_layers,
            "device": self.device,
        }

    @property
    def mlp_init_dict(self) -> Dict[str, Any]:
        """Returns dictionary of MLP information.

        :return: MLP information
        :rtype: Dict[str, Any]
        """
        base = self.base_init_dict.copy()
        extra_kwargs = {
            "hidden_size": self.hidden_size,
            "activation": self.activation,
            "output_activation": self.activation,
            "min_hidden_layers": self.min_hidden_layers,
            "max_hidden_layers": self.max_hidden_layers,
            "min_mlp_nodes": self.min_mlp_nodes,
            "max_mlp_nodes": self.max_mlp_nodes,
            "noisy": self.noisy,
            "noise_std": self.noise_std,
        }
        base.update(extra_kwargs)
        return base

    @property
    def cnn_init_dict(self) -> Dict[str, Any]:
        """Returns dictionary of CNN information.

        :return: CNN information
        :rtype: Dict[str, Any]
        """
        base = self.base_init_dict.copy()
        extra_kwargs = {
            "channel_size": self.channel_size,
            "kernel_size": self.kernel_size,
            "stride_size": self.stride_size,
            "activation": self.activation,
            "block_type": self.cnn_block_type,
            "output_activation": self.activation,
            "min_hidden_layers": self.min_cnn_hidden_layers,
            "max_hidden_layers": self.max_cnn_hidden_layers,
            "min_channel_size": self.min_channel_size,
            "max_channel_size": self.max_channel_size,
        }
        base.update(extra_kwargs)
        return base

    @property
    def init_dicts(self) -> Dict[str, Dict[str, Any]]:
        """Returns the initialization dictionaries for the network.

        :return: Initialization dictionaries
        :rtype: Dict[str, Dict[str, Any]]
        """
        if not hasattr(self, "feature_net"):
            return self._init_dicts

        reformatted_dicts = {}
        for key, net in self.feature_net.items():
            init_dict = net.init_dict
            init_dict.pop("input_shape", None)
            init_dict.pop("num_inputs", None)
            init_dict["num_outputs"] = self.latent_dim
            reformatted_dicts[key] = init_dict

        return reformatted_dicts

    def init_weights_gaussian(
        self, std_coeff: float = 4, output_coeff: float = 4
    ) -> None:
        """Initialise weights of linear layers using Gaussian distribution."""
        for module in self.feature_net.values():
            module.init_weights_gaussian(std_coeff=std_coeff)

        # Initialise final dense layer
        EvolvableModule.init_weights_gaussian(self.final_dense, std_coeff=output_coeff)

    def get_inner_init_dict(
        self, key: str, default: Literal["cnn", "mlp"]
    ) -> Dict[str, Any]:
        """Returns the initialization dictionary for the specified key.

        Arguments:
            key (str): Key of the observation space.

        Returns:
            Dict[str, Any]: Initialization dictionary.
        """
        init_dicts = self.init_dicts
        if key in init_dicts:
            return init_dicts[key]

        assert default in [
            "cnn",
            "mlp",
        ], "Invalid default value provided, must be 'cnn' or 'mlp'."

        return self.cnn_init_dict if default == "cnn" else self.mlp_init_dict

    def build_feature_extractor(self) -> Dict[str, SupportedEvolvableTypes]:
        """Creates the feature extractor and final MLP networks.

        :return: Dictionary of feature extractors.
        :rtype: Dict[str, Union[EvolvableCNN, EvolvableMLP]]
        """
        # Build feature extractors for image spaces only
        feature_net = ModuleDict(device=self.device)
        for i, (key, space) in enumerate(self.observation_space.spaces.items()):
            if is_image_space(space):  # Use CNN if it's an image space
                init_dict = copy.deepcopy(self.get_inner_init_dict(key, default="cnn"))

                if "sample_input" in init_dict:
                    sample_input = init_dict.pop("sample_input")
                else:
                    idx = i if self.is_tuple_space else key
                    sample_input = (
                        self.sample_input[idx]
                        if self.sample_input is not None
                        else None
                    )

                feature_extractor = EvolvableCNN(
                    input_shape=space.shape,
                    name=init_dict.pop("name", key),
                    sample_input=sample_input,
                    **init_dict
                )

                self._init_dicts[key] = feature_extractor.init_dict
                feature_net[key] = feature_extractor
            elif isinstance(space, spaces.Box) and len(space.shape) not in [0, 1]:
                raise ValueError(
                    "Box spaces with shape {} are not supported. Please use vector or image observations.".format(
                        space.shape
                    )
                )

        # Collect all vector space shapes for concatenation
        vector_input_dim = sum(
            [
                spaces.flatdim(self.observation_space.spaces[key])
                for key in self.vector_spaces
            ]
        )

        # Optional MLP for all concatenated vector inputs
        if self.vector_space_mlp:
            assert (
                self.hidden_size is not None
            ), "Hidden size must be specified for vector space MLP."

            init_dict = copy.deepcopy(
                self.get_inner_init_dict("vector_mlp", default="mlp")
            )
            vector_mlp = EvolvableMLP(
                num_inputs=vector_input_dim,
                name=init_dict.pop("name", "vector_mlp"),
                **init_dict
            )

            self._init_dicts["vector_mlp"] = vector_mlp.init_dict
            feature_net["vector_mlp"] = vector_mlp

        return feature_net

    def forward(self, x: TupleOrDictObservation) -> torch.Tensor:
        """Forward pass of the composed network. Extracts features from each observation key and concatenates
        them with the corresponding observation key if specified. The concatenated features are then passed
        through the final MLP to produce the output tensor.

        :param x: Dictionary of observations.
        :type x: Dict[str, ArrayOrTensor], Tuple[ArrayOrTensor]
        :param xc: Optional additional input tensor for critic network, defaults to None.
        :type xc: Optional[ArrayOrTensor], optional
        :param q: Flag to indicate if Q-values should be computed, defaults to True.
        :type q: bool, optional
        :return: Output tensor.
        :rtype: torch.Tensor
        """
        if isinstance(x, tuple):
            x = dict(zip(self.observation_space.spaces.keys(), x))

        for key, obs in x.items():
            if not isinstance(obs, torch.Tensor):
                x[key] = torch.tensor(obs, device=self.device, dtype=torch.float32)

        # Extract features from image spaces
        if self.image_features_dim > 0:
            image_features = [
                self.feature_net[key](x[key])
                for key in x.keys()
                if key in self.feature_net.keys()
            ]
            image_features = torch.cat(image_features, dim=1)

        # Handle case where there are no image subspaces
        else:
            image_features = torch.tensor([], device=self.device)

        # Extract raw features from vector spaces
        vector_inputs = []
        for key in self.vector_spaces:
            # Flatten if necessary
            if len(x[key].shape) > 2:
                x[key] = x[key].flatten(start_dim=1)
            elif len(x[key].shape) == 1:
                x[key] = x[key].unsqueeze(0)

            vector_inputs.append(x[key])

        # Concatenate vector inputs
        vector_inputs = torch.cat(vector_inputs, dim=1)

        # Pass through optional MLP
        if self.vector_space_mlp:
            vector_features = self.feature_net["vector_mlp"](vector_inputs)
        else:
            vector_features = vector_inputs

        # Concatenate all features and pass through final MLP
        features = torch.cat([image_features, vector_features], dim=1)
        return self.output(self.final_dense(features))

    @mutation(MutationType.ACTIVATION)
    def change_activation(self, activation: str, output: bool = False) -> None:
        """Set the activation function for the network.

        :param activation: Activation function to use.
        :type activation: str
        :param output: Flag indicating whether to set the output activation function, defaults to False
        :type output: bool, optional

        :return: Activation function
        :rtype: str
        """
        self.activation = activation
        if output:
            self.output = get_activation(activation)

    @mutation(MutationType.NODE)
    def add_latent_node(self, numb_new_nodes: Optional[int] = None) -> Dict[str, Any]:
        """Add a latent node to the network.

        :param numb_new_nodes: Number of new nodes to add, defaults to None
        :type numb_new_nodes: int, optional

        :return: Configuration for adding a latent node.
        :rtype: Dict[str, Any]
        """
        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([8, 16, 32], 1)[0]

        if self.latent_dim + numb_new_nodes < self.max_latent_dim:
            self.latent_dim += numb_new_nodes

        return {"numb_new_nodes": numb_new_nodes}

    @mutation(MutationType.NODE)
    def remove_latent_node(
        self, numb_new_nodes: Optional[int] = None
    ) -> Dict[str, Any]:
        """Remove a latent node from the network.

        :param numb_new_nodes: Number of nodes to remove, defaults to None
        :type numb_new_nodes: int, optional

        :return: Configuration for removing a latent node.
        :rtype: Dict[str, Any]
        """
        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([8, 16, 32], 1)[0]

        if self.latent_dim - numb_new_nodes > self.min_latent_dim:
            self.latent_dim -= numb_new_nodes

        return {"numb_new_nodes": numb_new_nodes}

    def recreate_network(self) -> None:
        """Recreates the network with the new latent dimension."""
        feature_net = self.build_feature_extractor()
        self.feature_net = EvolvableModule.preserve_parameters(
            old_net=self.feature_net, new_net=feature_net
        )

        # Collect all vector space shapes for concatenation
        vector_input_dim = sum(
            [
                spaces.flatdim(self.observation_space.spaces[key])
                for key in self.vector_spaces
            ]
        )

        # Calculate total feature dimension for final MLP
        image_features_dim = sum(
            [
                self.latent_dim
                for subspace in self.observation_space.spaces.values()
                if is_image_space(subspace)
            ]
        )

        vector_features_dim = (
            self.latent_dim if self.vector_space_mlp else vector_input_dim
        )
        features_dim = image_features_dim + vector_features_dim

        final_dense = nn.Linear(features_dim, self.num_outputs, device=self.device)
        self.final_dense = EvolvableModule.preserve_parameters(
            old_net=self.final_dense, new_net=final_dense
        )



import random
from collections import deque, namedtuple
from numbers import Number
from typing import Any, Deque, Dict, List, NamedTuple, Optional, Tuple, Union

import numpy as np
import torch
from numpy.typing import ArrayLike

from agilerl.components.segment_tree import MinSegmentTree, SumSegmentTree
from agilerl.typing import NumpyObsType
from agilerl.utils.algo_utils import obs_to_tensor

NpTransitionType = Union[Number, ArrayLike, Dict[str, ArrayLike]]
TorchTransitionType = Union[torch.Tensor, Dict[str, torch.Tensor]]


class ReplayBuffer:
    """The Experience Replay Buffer class. Used to store experiences and allow
    off-policy learning.

    :param memory_size: Maximum length of replay buffer
    :type memory_size: int
    :param field_names: Field names for experience named tuple, e.g. ['state', 'action', 'reward']
    :type field_names: list[str]
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to None
    :type device: str, optional
    """

    def __init__(
        self, memory_size: int, field_names: List[str], device: Optional[str] = None
    ):
        assert memory_size > 0, "Memory size must be greater than zero."
        assert len(field_names) > 0, "Field names must contain at least one field name."

        self.memory_size = memory_size
        self.memory = deque(maxlen=memory_size)
        self.field_names = field_names
        self.experience = namedtuple("Experience", field_names=self.field_names)
        self.counter = 0  # update cycle counter
        self.device = device

    def __len__(self) -> int:
        """Returns the current size of internal memory."""
        return len(self.memory)

    @staticmethod
    def stack_transitions(transitions: List[NumpyObsType]) -> NumpyObsType:
        """Stacks transitions into a single array/dictionary/tuple of arrays.

        :param transitions: List of transitions
        :type transitions: list[NumpyObsType]

        :return: Stacked transitions
        :rtype: NumpyObsType
        """
        # Identify the type of the transition
        field_type = type(transitions[0])

        # Stack the transitions into a single array or tuple/dictionary of arrays
        ts = []
        for ft in transitions:
            if field_type is dict:
                ts.append({k: np.expand_dims(v, axis=0) for k, v in ft.items()})
            elif field_type is tuple:
                ts.append(tuple(np.expand_dims(v, axis=0) for v in ft))
            else:
                ts.append(np.expand_dims(ft, axis=0))

        if field_type is dict:
            ts = {k: np.vstack([t[k] for t in ts]) for k in ts[0].keys()}
        elif field_type is tuple:
            ts = tuple(np.vstack([t[i] for t in ts]) for i in range(len(ts[0])))
        else:
            ts = np.vstack(ts)

        return ts

    def _add(self, *args: Any) -> None:
        """Adds experience to memory.

        :param *args: Variable length argument list. Contains transition elements in consistent order,
            e.g. state, action, reward, next_state, done
        """
        e = self.experience(*args)
        self.memory.append(e)

    def _process_transition(
        self, experiences: List[NamedTuple], np_array: bool = False
    ) -> Dict[str, Any]:
        """Returns transition dictionary from experiences.

        :param experiences: List of experiences
        :type experiences: list
        :param np_array: Flag to return numpy arrays instead of torch tensors, defaults to False
        :type np_array: bool, optional
        :return: Transition dictionary
        :rtype: dict
        """
        transition = {}
        for field in self.field_names:
            # Extract all of the transitions for the current field
            field_transitions: NpTransitionType = [
                getattr(e, field) for e in experiences if e is not None
            ]

            # Stack the transitions into a single array or tuple/dictionary of arrays
            ts = ReplayBuffer.stack_transitions(field_transitions)

            # Handle integer fields
            if field in [
                "done",
                "termination",
                "terminated",
                "truncation",
                "truncated",
            ]:
                ts = ts.astype(np.uint8)

            # Convert to torch tensor if specified
            if not np_array:
                ts = obs_to_tensor(ts, self.device)

            transition[field] = ts

        return transition

    def sample(
        self, batch_size: int, return_idx: bool = False, np_array: bool = False
    ) -> Tuple[Any, ...]:
        """Returns sample of experiences from memory.

        :param batch_size: Number of samples to return
        :type batch_size: int
        :param return_idx: Boolean flag to return index of samples randomly selected, defaults to False
        :type return_idx: bool, optional
        :param np_array: Flag to return numpy arrays instead of torch tensors, defaults to False
        :type np_array: bool, optional
        :return: Tuple of sampled experiences
        :rtype: tuple
        """
        if return_idx:
            idxs = np.random.choice(len(self.memory), size=batch_size, replace=False)
            experiences = list(map(lambda i: self.memory[i], idxs))
            transition = self._process_transition(experiences, np_array)
            transition["idxs"] = idxs
        else:
            experiences = random.sample(self.memory, k=batch_size)
            transition = self._process_transition(experiences, np_array)

        return tuple(transition.values())

    def save_to_memory_single_env(self, *args: Any) -> None:
        """Saves experience to memory.

        :param *args: Variable length argument list. Contains transition elements in consistent order,
            e.g. state, action, reward, next_state, done
        """
        self._add(*args)
        self.counter += 1

    def save_to_memory_vect_envs(self, *args: Any) -> None:
        """Saves multiple experiences to memory.

        :param *args: Variable length argument list. Contains batched transition elements in consistent order,
            e.g. states, actions, rewards, next_states, dones
        """
        for transition in zip(*args):
            self._add(*transition)
            self.counter += 1

    def save_to_memory(self, *args: Any, is_vectorised: bool = False) -> None:
        """Applies appropriate save_to_memory function depending on whether
        the environment is vectorised or not.

        :param *args: Variable length argument list. Contains batched or unbatched transition elements in consistent order,
            e.g. states, actions, rewards, next_states, dones
        :param is_vectorised: Boolean flag indicating if the environment has been vectorised
        :type is_vectorised: bool
        """
        if is_vectorised:
            self.save_to_memory_vect_envs(*args)
        else:
            self.save_to_memory_single_env(*args)


class MultiStepReplayBuffer(ReplayBuffer):
    """The Multi-step Experience Replay Buffer class. Used to store experiences and allow
    off-policy learning.

    :param memory_size: Maximum length of replay buffer
    :type memory_size: int
    :param field_names: Field names for experience named tuple, e.g. ['state', 'action', 'reward']
    :type field_names: list[str]
    :param num_envs: Number of parallel environments for training
    :type num_envs: int
    :param n_step: Step number to calculate n-step td error, defaults to 3
    :type n_step: int, optional
    :param gamma: Discount factor, defaults to 0.99
    :type gamma: float, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to None
    :type device: str, optional
    """

    def __init__(
        self,
        memory_size: int,
        field_names: List[str],
        num_envs: int,
        n_step: int = 3,
        gamma: float = 0.99,
        device: Optional[str] = None,
    ):
        super().__init__(memory_size, field_names, device)
        assert (
            "reward" in field_names
        ), "Reward must be saved in replay buffer under the field name 'reward'."
        assert (
            "next_state" in field_names
        ), "Next state must be saved in replay buffer under the field name 'next_state'."
        assert (
            "done" in field_names
            or "termination" in field_names
            or "terminated" in field_names
        ), "Done/termination must be saved in replay buffer under the field name 'done', 'termination', or 'terminated'."

        self.num_envs = num_envs
        self.n_step_buffers = [deque(maxlen=n_step) for _ in range(num_envs)]
        self.args_deque = deque(maxlen=n_step)
        self.n_step = n_step
        self.gamma = gamma

    def save_to_memory_single_env(self, *args: Any) -> Tuple[Any, ...]:
        """Saves experience to memory.

        :param *args: Variable length argument list. Contains transition elements in consistent order,
            e.g. state, action, reward, next_state, done
        :return: The first saved transition
        :rtype: tuple
        """
        self.args_deque.append(args)
        transition = self.experience(*args)
        self.n_step_buffers[0].append(transition)

        # single step transition is not ready
        if len(self.n_step_buffers[0]) < self.n_step:
            return ()

        # make a n-step transition
        args = self._get_n_step_info(self.n_step_buffers[0], self.gamma)
        self._add(*args)
        self.counter += 1

        return self.args_deque[0]

    def save_to_memory_vect_envs(self, *args: Any) -> Tuple[Any, ...]:
        """Saves multiple experiences to memory.

        :param *args: Variable length argument list. Contains transition elements in consistent order,
            e.g. state, action, reward, next_state, done
        :return: The first saved transition
        :rtype: tuple
        """
        self.args_deque.append(args)
        for buffer, *transition in zip(self.n_step_buffers, *args):
            transition = self.experience(*transition)
            buffer.append(transition)

        # single step transition is not ready
        if any(len(buffer) < self.n_step for buffer in self.n_step_buffers):
            return ()
        else:
            for buffer in self.n_step_buffers:
                # make a n-step transition
                single_step_args = self._get_n_step_info(buffer, self.gamma)
                self._add(*single_step_args)
                self.counter += 1

            return self.args_deque[0]

    def sample_from_indices(self, idxs: List[int]) -> Tuple[Any, ...]:
        """Returns sample of experiences from memory using provided indices.

        :param idxs: Indices to sample
        :type idxs: list[int]
        :return: Tuple of sampled experiences
        :rtype: tuple
        """
        experiences = list(map(lambda i: self.memory[i], idxs))
        transition = self._process_transition(experiences)
        return tuple(transition.values())

    def _get_n_step_info(
        self, n_step_buffer: Deque[NamedTuple], gamma: float
    ) -> Tuple[Any, ...]:
        """Returns n step reward, next_state, and done, as well as other saved transition elements, in order.

        :param n_step_buffer: Buffer containing n-step transitions
        :type n_step_buffer: deque
        :param gamma: Discount factor
        :type gamma: float
        :return: Tuple containing n-step transition elements
        :rtype: tuple
        """
        # info of the last transition
        t = [n_step_buffer[0]]
        transition = self._process_transition(t, np_array=True)

        vect_reward = transition["reward"][0]
        vect_next_state = transition["next_state"][0]
        if "done" in transition.keys():
            vect_done = transition["done"][0]
        elif "termination" in transition.keys():
            vect_done = transition["termination"][0]
        else:
            vect_done = transition["terminated"][0]

        for idx, ts in enumerate(list(n_step_buffer)[1:]):
            if not vect_done:
                vect_r, vect_n_s = (ts.reward, ts.next_state)

                if "done" in transition.keys():
                    vect_d = ts.done
                elif "termination" in transition.keys():
                    vect_d = ts.termination
                else:
                    vect_d = ts.terminated

                vect_reward += vect_r * gamma ** (idx + 1)
                vect_done = np.array([vect_d])
                vect_next_state = vect_n_s

        transition["reward"] = vect_reward
        transition["next_state"] = vect_next_state
        if "done" in transition.keys():
            transition["done"] = vect_done
        elif "termination" in transition.keys():
            transition["termination"] = vect_done
        else:
            transition["terminated"] = vect_done
        transition["state"] = transition["state"][0]
        transition["action"] = transition["action"][0]

        return tuple(transition.values())


class PrioritizedReplayBuffer(MultiStepReplayBuffer):
    """The Prioritized Experience Replay Buffer class. Used to store experiences and allow
    off-policy learning.

    :param memory_size: Maximum length of replay buffer
    :type memory_size: int
    :param field_names: Field names for experience named tuple, e.g. ['state', 'action', 'reward']
    :type field_names: list[str]
    :param num_envs: Number of parallel environments for training
    :type num_envs: int
    :param alpha: Alpha parameter for prioritized replay buffer, defaults to 0.6
    :type alpha: float, optional
    :param n_step: Step number to calculate n-step td error, defaults to 1
    :type n_step: int, optional
    :param gamma: Discount factor, defaults to 0.99
    :type gamma: float, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to None
    :type device: str, optional
    """

    def __init__(
        self,
        memory_size: int,
        field_names: List[str],
        num_envs: int,
        alpha: float = 0.6,
        n_step: int = 1,
        gamma: float = 0.99,
        device: Optional[str] = None,
    ):
        super().__init__(memory_size, field_names, num_envs, n_step, gamma, device)
        self.max_priority, self.tree_ptr = 1.0, 0
        self.alpha = alpha

        # capacity must be positive and a power of 2.
        tree_capacity = 1
        while tree_capacity < self.memory_size:
            tree_capacity *= 2

        self.sum_tree = SumSegmentTree(tree_capacity)
        self.min_tree = MinSegmentTree(tree_capacity)

    def _add(self, *args: Any) -> None:
        """Adds experience to memory and updates priority trees.

        :param *args: Variable length argument list. Contains transition elements in consistent order,
            e.g. state, action, reward, next_state, done
        """
        super()._add(*args)
        self.sum_tree[self.tree_ptr] = self.max_priority**self.alpha
        self.min_tree[self.tree_ptr] = self.max_priority**self.alpha
        self.tree_ptr = (self.tree_ptr + 1) % self.memory_size

    def sample(self, batch_size: int, beta: float = 0.4) -> Tuple[Any, ...]:
        """Returns sample of experiences from memory.

        :param batch_size: Number of samples to return
        :type batch_size: int
        :param beta: Beta parameter for importance sampling, defaults to 0.4
        :type beta: float, optional
        :return: Tuple of sampled experiences
        :rtype: tuple
        """
        idxs = self._sample_proportional(batch_size)
        experiences = [self.memory[i] for i in idxs]
        transition = self._process_transition(experiences)

        weights = torch.from_numpy(
            np.array([self._calculate_weight(i, beta) for i in idxs])
        ).float()

        if self.device is not None:
            weights = weights.to(self.device)

        transition["weights"] = weights
        transition["idxs"] = idxs

        return tuple(transition.values())

    def update_priorities(self, idxs: List[int], priorities: List[float]) -> None:
        """Update priorities of sampled transitions.

        :param idxs: Indices of sampled transitions
        :type idxs: list[int]
        :param priorities: New priorities of sampled transitions
        :type priorities: list[float]
        """
        for idx, priority in zip(idxs, priorities):
            self.sum_tree[idx] = priority**self.alpha
            self.min_tree[idx] = priority**self.alpha
            self.max_priority = max(self.max_priority, priority)

    def _sample_proportional(self, batch_size: int) -> List[int]:
        """Sample indices based on proportions.

        :param batch_size: Sample size
        :type batch_size: int
        :return: List of sampled indices
        :rtype: list[int]
        """
        idxs = []
        p_total = self.sum_tree.sum(0, len(self) - 1)
        segment = p_total / batch_size
        for i in range(batch_size):
            a = segment * i
            b = segment * (i + 1)
            upperbound = random.uniform(a, b)
            idx = self.sum_tree.retrieve(upperbound)
            idxs.append(idx)
        return idxs

    def _calculate_weight(self, idx: int, beta: float) -> float:
        """Calculate the weight of the experience at idx.

        :param idx: Index of the experience
        :type idx: int
        :param beta: Beta parameter for importance sampling
        :type beta: float
        :return: Weight of the experience
        :rtype: float
        """
        # get max weight
        p_min = self.min_tree.min() / self.sum_tree.sum()
        max_weight = (p_min * len(self)) ** (-beta)

        # calculate weights
        p_sample = self.sum_tree[idx] / self.sum_tree.sum()
        weight = (p_sample * len(self)) ** (-beta)
        weight = weight / max_weight

        return weight



from .base import EvolvableAlgorithm, MultiAgentRLAlgorithm, RLAlgorithm
from .wrappers import OptimizerWrapper

__all__ = [
    "EvolvableAlgorithm",
    "RLAlgorithm",
    "MultiAgentRLAlgorithm",
    "OptimizerWrapper",
]



from typing import Optional, Type, Union

import numpy as np
import torch
from gymnasium import spaces
from torch.distributions import Bernoulli, Categorical, Distribution, Normal

from agilerl.modules.base import EvolvableModule, EvolvableWrapper
from agilerl.modules.configs import MlpNetConfig
from agilerl.networks.base import EvolvableNetwork
from agilerl.typing import ArrayOrTensor, ConfigType, DeviceType, TorchObsType


class EvolvableDistribution(EvolvableWrapper):
    """Wrapper to output a distribution over an action space for an evolvable module. It provides methods
    to sample actions and compute log probabilities, relevant for many policy-gradient algorithms such as
    PPO, A2C, TRPO.

    :param action_space: Action space of the environment.
    :type action_space: spaces.Space
    :param network: Network that outputs the logits of the distribution.
    :type network: EvolvableModule
    """

    wrapped: EvolvableModule

    def __init__(
        self,
        action_space: spaces.Space,
        network: EvolvableModule,
        log_std_init: float = 0.0,
        device: DeviceType = "cpu",
    ):

        super().__init__(network)

        self.action_space = action_space
        self.log_std_init = log_std_init
        self.device = device

        # For continuous action spaces, we also learn the standard deviation (log_std)
        # of the action distribution
        if isinstance(action_space, spaces.Box):
            self.log_std = torch.nn.Parameter(
                torch.ones(spaces.flatdim(action_space)) * log_std_init,
                requires_grad=True,
            ).to(device)
        else:
            self.log_std = None

    @property
    def net_config(self) -> ConfigType:
        """Configuration of the network.

        :return: Configuration of the network.
        :rtype: ConfigType
        """
        return self.wrapped.net_config

    def get_distribution(
        self, probs: torch.Tensor, log_std: Optional[torch.Tensor] = None
    ) -> Distribution:
        """Get the distribution over the action space given an observation.

        :param probs: Logits output by the network.
        :type probs: torch.Tensor
        :param log_std: Log standard deviation of the action distribution. Defaults to None.
        :type log_std: Optional[torch.Tensor]
        :return: Distribution over the action space.
        :rtype: Distribution
        """
        if isinstance(self.action_space, spaces.Box):
            assert (
                log_std is not None
            ), "log_std must be provided for continuous action spaces."
            std = torch.ones_like(probs) * log_std.exp()
            return Normal(loc=probs, scale=std)

        elif isinstance(self.action_space, spaces.Discrete):
            return Categorical(probs=probs)

        elif isinstance(self.action_space, spaces.MultiDiscrete):
            return [
                Categorical(probs=split)
                for split in torch.split(probs, list(self.action_space.nvec), dim=1)
            ]

        elif isinstance(self.action_space, spaces.MultiBinary):
            return Bernoulli(probs=probs)

        else:
            raise NotImplementedError(
                f"Action space {self.action_space} not supported."
            )

    def forward(
        self, latent: torch.Tensor, action_mask: Optional[ArrayOrTensor] = None
    ) -> Distribution:
        """Forward pass of the network.

        :param obs: Observation input.
        :type obs: TorchObsType
        :param action_mask: Mask to apply to the logits. Defaults to None.
        :type action_mask: Optional[ArrayOrTensor]
        :return: Distribution over the action space.
        :rtype: Distribution
        """
        probs = self.wrapped(latent)

        if action_mask is not None and isinstance(self.action_space, spaces.Discrete):
            # Safe to assume the action masks will always be NumPy arrays
            if isinstance(action_mask, np.ndarray):
                action_mask = (
                    np.stack(action_mask)
                    if action_mask.dtype == np.object_ or isinstance(action_mask, list)
                    else action_mask
                )
            action_mask = torch.as_tensor(
                action_mask, dtype=torch.bool, device=self.device
            ).reshape(probs.shape)
            probs = torch.where(
                action_mask,
                probs,
                torch.tensor(0.0, dtype=probs.dtype, device=self.device),
            )

        return self.get_distribution(probs, self.log_std)

    def clone(self) -> "EvolvableDistribution":
        """Clones the distribution.

        :return: Cloned distribution.
        :rtype: EvolvableDistribution
        """
        return EvolvableDistribution(
            action_space=self.action_space,
            network=self.wrapped.clone(),
            log_std_init=self.log_std_init,
            device=self.device,
        )


class DeterministicActor(EvolvableNetwork):
    """Deterministic actor network for policy-gradient algorithms. Given an observation, it outputs
    the mean of the action distribution. This is useful for e.g. DDPG, SAC, TD3.

    :param observation_space: Observation space of the environment.
    :type observation_space: spaces.Space
    :param action_space: Action space of the environment
    :type action_space: spaces.Space
    :param encoder_cls: Encoder class to use for the network. Defaults to None, whereby it is
        automatically built using an AgileRL module according the observation space.
    :type encoder_cls: Optional[Union[str, Type[EvolvableModule]]]
    :param encoder_config: Configuration of the encoder network.
    :type encoder_config: ConfigType
    :param head_config: Configuration of the network MLP head.
    :type head_config: Optional[ConfigType]
    :param min_latent_dim: Minimum dimension of the latent space representation.
    :type min_latent_dim: int
    :param max_latent_dim: Maximum dimension of the latent space representation.
    :type max_latent_dim: int
    :param n_agents: Number of agents in the environment. Defaults to None, which corresponds to
        single-agent environments.
    :type n_agents: Optional[int]
    :param latent_dim: Dimension of the latent space representation.
    :type latent_dim: int
    :param device: Device to use for the network.
    :type device: str
    """

    def __init__(
        self,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        encoder_cls: Optional[Union[str, Type[EvolvableModule]]] = None,
        encoder_config: Optional[ConfigType] = None,
        head_config: Optional[ConfigType] = None,
        min_latent_dim: int = 8,
        max_latent_dim: int = 128,
        n_agents: Optional[int] = None,
        latent_dim: int = 32,
        simba: bool = False,
        device: str = "cpu",
    ):

        super().__init__(
            observation_space,
            encoder_cls=encoder_cls,
            encoder_config=encoder_config,
            action_space=action_space,
            min_latent_dim=min_latent_dim,
            max_latent_dim=max_latent_dim,
            n_agents=n_agents,
            latent_dim=latent_dim,
            simba=simba,
            device=device,
        )

        self.min_action = (
            action_space.low if isinstance(action_space, spaces.Box) else None
        )
        self.max_action = (
            action_space.high if isinstance(action_space, spaces.Box) else None
        )

        # Set output activation based on action space
        if isinstance(action_space, (spaces.Discrete, spaces.MultiDiscrete)):
            output_activation = "Softmax"
        elif np.any(self.min_action < 0):
            output_activation = "Tanh"
        else:
            output_activation = "Sigmoid"

        if head_config is None:
            head_config = MlpNetConfig(
                hidden_size=[16], output_activation=output_activation
            )
        elif head_config.get("output_activation", None) is None:
            head_config["output_activation"] = output_activation

        self.build_network_head(head_config)
        self.output_activation = head_config.get("output_activation", output_activation)

    def build_network_head(self, net_config: Optional[ConfigType] = None) -> None:
        """Builds the head of the network.

        :param net_config: Configuration of the head.
        :type net_config: Optional[ConfigType]
        """
        self.head_net = self.create_mlp(
            num_inputs=self.latent_dim,
            num_outputs=spaces.flatdim(self.action_space),
            name="actor",
            net_config=net_config,
        )

    def forward(self, obs: TorchObsType) -> torch.Tensor:
        """Forward pass of the network.

        :param obs: Observation input.
        :type obs: TorchObsType
        :return: Output of the network.
        :rtype: torch.Tensor
        """
        latent = self.encoder(obs)
        return self.head_net(latent)

    def recreate_network(self) -> None:
        """Recreates the network."""
        self.recreate_encoder()

        head_net = self.create_mlp(
            num_inputs=self.latent_dim,
            num_outputs=spaces.flatdim(self.action_space),
            name="actor",
            net_config=self.head_net.net_config,
        )

        self.head_net = EvolvableModule.preserve_parameters(self.head_net, head_net)


class StochasticActor(DeterministicActor):
    """Stochastic actor network for policy-gradient algorithms. Given an observation, it outputs
    a distribution over the action space. This is useful for on-policy policy-gradient algorithms
    like PPO, A2C, TRPO.

    :param observation_space: Observation space of the environment.
    :type observation_space: spaces.Space
    :param action_space: Action space of the environment
    :type action_space: spaces.Space
    :param encoder_cls: Encoder class to use for the network. Defaults to None, whereby it is
        automatically built using an AgileRL module according the observation space.
    :type encoder_cls: Optional[Union[str, Type[EvolvableModule]]]
    :param encoder_config: Configuration of the encoder network.
    :type encoder_config: ConfigType
    :param head_config: Configuration of the network MLP head.
    :type head_config: Optional[ConfigType]
    :param n_agents: Number of agents in the environment. Defaults to None, which corresponds to
        single-agent environments.
    :type n_agents: Optional[int]
    :param latent_dim: Dimension of the latent space representation.
    :type latent_dim: int
    :param device: Device to use for the network.
    :type device: str
    """

    head_net: EvolvableDistribution

    def __init__(
        self,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        encoder_cls: Optional[Union[str, Type[EvolvableModule]]] = None,
        encoder_config: Optional[ConfigType] = None,
        head_config: Optional[ConfigType] = None,
        log_std_init: float = 0.0,
        min_latent_dim: int = 8,
        max_latent_dim: int = 128,
        n_agents: Optional[int] = None,
        latent_dim: int = 32,
        simba: bool = False,
        device: str = "cpu",
    ):

        super().__init__(
            observation_space,
            action_space=action_space,
            encoder_cls=encoder_cls,
            encoder_config=encoder_config,
            head_config=head_config,
            min_latent_dim=min_latent_dim,
            max_latent_dim=max_latent_dim,
            n_agents=n_agents,
            latent_dim=latent_dim,
            simba=simba,
            device=device,
        )

        self.log_std_init = log_std_init
        self.head_net = EvolvableDistribution(
            action_space, self.head_net, log_std_init=log_std_init, device=device
        )

    def forward(
        self, obs: TorchObsType, action_mask: Optional[ArrayOrTensor] = None
    ) -> Distribution:
        """Forward pass of the network.

        :param obs: Observation input.
        :type obs: TorchObsType
        :return: Distribution over the action space.
        :rtype: Distribution
        """
        latent = self.encoder(obs)
        return self.head_net.forward(latent, action_mask)

    def __call__(
        self, obs: TorchObsType, action_mask: Optional[ArrayOrTensor] = None
    ) -> Distribution:
        """Calls the forward method.

        :param obs: Observation input.
        :type obs: TorchObsType
        :return: Distribution over the action space.
        :rtype: Distribution
        """
        return self.forward(obs, action_mask)

    def recreate_network(self) -> None:
        """Recreates the network with the same parameters as the current network.

        :param shrink_params: Whether to shrink the parameters of the network. Defaults to False.
        :type shrink_params: bool
        """
        self.recreate_encoder()

        head_net = self.create_mlp(
            num_inputs=self.latent_dim,
            num_outputs=spaces.flatdim(self.action_space),
            name="actor",
            net_config=self.head_net.net_config,
        )

        head_net = EvolvableDistribution(
            self.action_space,
            head_net,
            log_std_init=self.log_std_init,
            device=self.device,
        )

        self.head_net = EvolvableModule.preserve_parameters(self.head_net, head_net)



from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import torch
import yaml


@dataclass
class NetConfig:
    """Dataclass for storing evolvable network configurations."""

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> "NetConfig":
        return cls(**config)

    @classmethod
    def from_yaml(cls, path: str) -> "NetConfig":
        with open(path) as file:
            config: Dict[str, Any] = yaml.safe_load(file)
            assert "NET_CONFIG" in config, "NET_CONFIG not found in yaml file."
            net_config = config["NET_CONFIG"]

        return cls.from_dict(net_config)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:
        setattr(self, key, value)

    def get(self, key: str, default: Any = None) -> Any:
        return getattr(self, key, default)

    def pop(self, key: str, default: Any = None) -> Any:
        attr = getattr(self, key, default)
        delattr(self, key)
        return attr

    def keys(self) -> List[str]:
        return list(self.__dataclass_fields__.keys())

    def values(self) -> List[Any]:
        return [getattr(self, key) for key in self.keys()]

    def items(self) -> Dict[str, Any]:
        return {key: getattr(self, key) for key in self.keys()}.items()


@dataclass
class MlpNetConfig(NetConfig):
    hidden_size: List[int]
    activation: str = field(default="ReLU")
    output_activation: Optional[str] = field(default=None)
    min_hidden_layers: int = field(default=1)
    max_hidden_layers: int = field(default=3)
    min_mlp_nodes: int = field(default=16)
    max_mlp_nodes: int = field(default=500)
    layer_norm: bool = field(default=True)
    output_vanish: bool = field(default=True)
    init_layers: bool = field(default=True)
    noisy: bool = field(default=False)
    noise_std: float = field(default=0.5)

    def __post_init__(self):
        assert (
            len(self.hidden_size) >= self.min_hidden_layers
        ), "Hidden layers must be greater than min_hidden_layers."

        assert (
            len(self.hidden_size) <= self.max_hidden_layers
        ), "Hidden layers must be less than max_hidden_layers."

        assert all(
            [
                self.min_mlp_nodes <= nodes and nodes <= self.max_mlp_nodes
                for nodes in self.hidden_size
            ]
        ), "Nodes must be within min_nodes and max_nodes."


@dataclass
class SimBaNetConfig(NetConfig):
    hidden_size: int
    num_blocks: int
    output_activation: Optional[str] = field(default=None)
    min_blocks: int = field(default=1)
    max_blocks: int = field(default=4)
    min_mlp_nodes: int = field(default=16)
    max_mlp_nodes: int = field(default=500)

    def __post_init__(self):
        assert (
            self.num_blocks >= self.min_blocks
        ), "Number of residual blocks must be greater than min_blocks."

        assert (
            self.num_blocks <= self.max_blocks
        ), "Number of residual blocks must be less than max_blocks."

        assert (
            self.min_mlp_nodes <= self.hidden_size
            and self.hidden_size <= self.max_mlp_nodes
        ), "Nodes must be within min_nodes and max_nodes."


@dataclass
class CnnNetConfig(NetConfig):
    channel_size: List[int]
    kernel_size: List[Union[int, Tuple[int, ...]]]
    stride_size: List[int]
    sample_input: Optional[torch.Tensor] = field(default=None)
    activation: str = field(default="ReLU")
    output_activation: Optional[str] = field(default=None)
    block_type: Literal["Conv2d", "Conv3d"] = field(default="Conv2d")
    min_hidden_layers: int = field(default=1)
    max_hidden_layers: int = field(default=6)
    min_channel_size: int = field(default=32)
    max_channel_size: int = field(default=256)
    layer_norm: bool = field(default=True)
    init_layers: bool = field(default=True)


@dataclass
class MultiInputNetConfig(NetConfig):
    channel_size: List[int]
    kernel_size: List[int]
    stride_size: List[int]
    latent_dim: int = 16
    cnn_block_type: Literal["Conv2d", "Conv3d"] = "Conv2d"
    sample_input: Optional[torch.Tensor] = field(default=None)
    vector_space_mlp: bool = False
    hidden_size: Optional[List[int]] = field(default=None)
    init_dicts: Dict[str, Dict[str, Any]] = field(default_factory=dict)
    activation: str = "ReLU"
    output_activation: Optional[str] = field(default=None)
    min_hidden_layers: int = 1
    max_hidden_layers: int = 3
    min_mlp_nodes: int = 64
    max_mlp_nodes: int = 1024
    min_cnn_hidden_layers: int = 1
    max_cnn_hidden_layers: int = 6
    min_channel_size: int = 32
    max_channel_size: int = 256
    min_latent_dim: int = 8
    max_latent_dim: int = 128
    layer_norm: bool = False
    noisy: bool = False
    noise_std: float = 0.5
    init_layers: bool = True



from typing import Any, ClassVar, Dict, List, Protocol, Tuple, Union

import gymnasium as gym
import numpy as np
import torch
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces
from numpy.typing import ArrayLike
from torch.nn import Module
from torch.optim import Optimizer

from agilerl.protocols import EvolvableAlgorithm


class IsDataclass(Protocol):
    __dataclass_fields__: ClassVar[Dict[str, Any]]


ArrayOrTensor = Union[ArrayLike, torch.Tensor]
TensorDict = Dict[str, torch.Tensor]
TensorTuple = Tuple[torch.Tensor, ...]
ArrayDict = Dict[str, np.ndarray]
ArrayTuple = Tuple[ArrayLike, ...]
NetConfigType = Dict[str, Any]
KernelSizeType = Union[int, Tuple[int, ...]]
SupportedGymSpaces = Union[
    spaces.Box, spaces.Discrete, spaces.MultiDiscrete, spaces.Dict, spaces.Tuple
]
GymSpaceType = Union[SupportedGymSpaces, List[SupportedGymSpaces]]
GymEnvType = Union[str, gym.Env, gym.vector.VectorEnv]

NumpyObsType = Union[np.ndarray, ArrayDict, ArrayTuple]
TorchObsType = Union[torch.Tensor, TensorDict, TensorTuple]
ObservationType = Union[NumpyObsType, TorchObsType]
ActionType = Union[int, float, ArrayLike, torch.Tensor]
InfosDict = Dict[str, Dict[str, Any]]
MaybeObsList = Union[List[ObservationType], ObservationType]
ExperiencesType = Tuple[ObservationType, ...]
StepType = Tuple[NumpyObsType, ActionType, float, MaybeObsList, InfosDict]
MultiAgentStepType = Tuple[
    Dict[str, NumpyObsType], ArrayDict, ArrayDict, ArrayDict, Dict[str, Any]
]

DeviceType = Union[str, torch.device]
OptimizerType = Union[Optimizer, AcceleratedOptimizer]
NetworkType = Union[Module, List[Module], Tuple[Module, ...]]
PopulationType = List[EvolvableAlgorithm]
ConfigType = Union[IsDataclass, Dict[str, Any]]
StateDict = Union[Dict[str, Any], List[Dict[str, Any]]]



import copy
import inspect
import warnings
from collections import OrderedDict, defaultdict
from numbers import Number
from typing import Any, Dict, Iterable, List, Optional, Tuple, TypeGuard, Union

import numpy as np
import torch
import torch.nn.functional as F
from accelerate.optimizer import AcceleratedOptimizer
from accelerate.utils.deepspeed import DeepSpeedOptimizerWrapper
from gymnasium import spaces
from tensordict.nn import CudaGraphModule
from torch._dynamo import OptimizedModule
from torch.nn import Module
from torch.optim import Optimizer

from agilerl.protocols import EvolvableAttributeType, EvolvableModule, OptimizerWrapper
from agilerl.typing import (
    ArrayOrTensor,
    MaybeObsList,
    NetworkType,
    NumpyObsType,
    ObservationType,
    OptimizerType,
    TorchObsType,
)


def is_image_space(space: spaces.Space) -> bool:
    """Check if the space is an image space. We ignore dtype and number of channels
    checks.

    :param space: Input space
    :type space: spaces.Space

    :return: True if the space is an image space, False otherwise
    :rtype: bool
    """
    return isinstance(space, spaces.Box) and len(space.shape) == 3


def contains_image_space(space: spaces.Space) -> bool:
    """Checks if the space contains an image space.

    :param space: Observation space
    :type space: spaces.Space
    :return: True if the space contains an image space, False otherwise
    :rtype: bool
    """
    if isinstance(space, spaces.Dict):
        return any(contains_image_space(subspace) for subspace in space.spaces.values())
    elif isinstance(space, spaces.Tuple):
        return any(contains_image_space(subspace) for subspace in space.spaces)
    elif isinstance(space, spaces.Box):
        return is_image_space(space)
    return False


def multi_agent_sample_tensor_from_space(
    space: spaces.Space,
    n_agents: int,
    critic: bool = False,
    device: torch.device = torch.device("cpu"),
) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
    """Gets the sample tensor from an observation space for multi-agent settings.

    :param space: Observation space
    :type space: spaces.Space
    :param n_agents: Number of agents
    :type n_agents: int
    :param critic: If True, the tensor is for the critic, defaults to False
    :type critic: bool, optional

    :return: Sample tensor
    :rtype: torch.Tensor or dict[str, torch.Tensor]
    """

    def sample(image_shape: Tuple[int, ...], critic: bool) -> torch.Tensor:
        tensor = torch.zeros(
            (1, *image_shape), dtype=torch.float32, device=device
        ).unsqueeze(2)

        if critic:
            tensor = tensor.repeat(1, 1, n_agents, 1, 1)

        return tensor

    if isinstance(space, spaces.Dict):
        sample_tensor = {
            key: multi_agent_sample_tensor_from_space(
                subspace, n_agents, critic, device
            )
            for key, subspace in space.spaces.items()
            if is_image_space(subspace)
        }
    elif isinstance(space, spaces.Tuple):
        sample_tensor = tuple(
            (
                multi_agent_sample_tensor_from_space(subspace, n_agents, critic, device)
                if is_image_space(subspace)
                else None
            )
            for subspace in space.spaces
        )
    elif is_image_space(space):
        sample_tensor = sample(space.shape, critic)
    else:
        sample_tensor = None

    return sample_tensor


def make_safe_deepcopies(*args: Union[Module, List[Module]]) -> List[Module]:
    """Makes deep copies of EvolvableModule objects and their attributes.

    :param args: EvolvableModule or lists of EvolvableModule objects to copy.
    :type args: Union[EvolvableModule, List[EvolvableModule]].

    :return: Deep copies of the EvolvableModule objects and their attributes.
    :rtype: List[EvolvableModule].
    """
    copies = []
    for arg in args:
        if isinstance(arg, list):
            arg_copy = [
                copy.deepcopy(inner_arg.cpu()).to(inner_arg.device) for inner_arg in arg
            ]
        else:
            arg_copy = copy.deepcopy(arg.cpu()).to(arg.device)

        copies.append(arg_copy)

    return copies[0] if len(copies) == 1 else copies


def is_module_list(obj: EvolvableAttributeType) -> TypeGuard[Iterable[EvolvableModule]]:
    """Type guard to check if an object is a list of EvolvableModule objects.

    :param obj: The object to check.
    :type obj: EvolvableAttributeType.

    :return: True if the object is a list of EvolvableModule objects, False otherwise.
    :rtype: bool.
    """
    if not isinstance(obj, list):
        return False

    return all(
        isinstance(inner_obj, (OptimizedModule, EvolvableModule)) for inner_obj in obj
    )


def is_optimizer_list(obj: EvolvableAttributeType) -> TypeGuard[Iterable[Optimizer]]:
    """Type guard to check if an object is a list of Optimizer's.

    :param obj: The object to check.
    :type obj: EvolvableAttributeType.

    :return: True if the object is a list of Optimizer's, False otherwise.
    :rtype: bool.
    """
    return all(isinstance(inner_obj, Optimizer) for inner_obj in obj)


def assert_supported_space(space: spaces.Space) -> bool:
    """Checks if the space is supported by the AgileRL framework.

    :param space: The space to check.
    :type space: spaces.Space.

    :return: True if the space is supported, False otherwise.
    :rtype: bool.
    """
    # Nested Dict or Tuple spaces are not supported
    if isinstance(space, spaces.Dict) and any(
        isinstance(subspace, (spaces.Dict, spaces.Tuple))
        for subspace in space.spaces.values()
    ):
        raise TypeError(f"Nested {type(space)} spaces are not supported.")
    elif isinstance(space, spaces.Tuple) and any(
        isinstance(subspace, (spaces.Dict, spaces.Tuple)) for subspace in space.spaces
    ):
        raise TypeError(f"Nested {type(space)} spaces are not supported.")


def isroutine(obj: object) -> bool:
    """Checks if an attribute is a routine, considering also methods wrapped by
    CudaGraphModule.

    :param attr: The attribute to check.
    :type attr: str

    :return: True if the attribute is a routine, False otherwise.
    :rtype: bool
    """
    if isinstance(obj, CudaGraphModule):
        return True

    return inspect.isroutine(obj)


def unwrap_optimizer(
    optimizer: OptimizerType, network: NetworkType, lr: float
) -> Optimizer:
    """Unwraps AcceleratedOptimizer to get the underlying optimizer.

    :param optimizer: AcceleratedOptimizer
    :type optimizer: AcceleratedOptimizer
    :param network: Network or list of networks
    :type network: Union[Module, List[Module], Tuple[Module, ...]]
    :param lr: Learning rate
    :type lr: float
    :return: Unwrapped optimizer
    rtype: Optimizer
    """
    if isinstance(optimizer, AcceleratedOptimizer):
        if isinstance(network, (list, tuple)):
            optim_arg = [{"params": net.parameters(), "lr": lr} for net in network]
            unwrapped_optimizer: Optimizer = type(optimizer.optimizer)(optim_arg)
        else:
            unwrapped_optimizer: Optimizer = type(optimizer.optimizer)(
                network.parameters(), lr=lr
            )
        unwrapped_optimizer.load_state_dict(optimizer.state_dict())
        return unwrapped_optimizer
    else:
        return optimizer


def recursive_check_module_attrs(obj: Any, networks_only: bool = False) -> bool:
    """Recursively check if the object has any attributes that are EvolvableModule objects or Optimizer's.

    :param obj: The object to check for EvolvableModule objects or Optimizer's.
    :type obj: Any
    :return: True if the object has any attributes that are EvolvableModule objects or Optimizer's, False otherwise.
    :rtype: bool
    :param networks_only: If True, only check for EvolvableModule objects, defaults to False
    :type networks_only: bool, optional
    """
    check_types = (OptimizedModule, EvolvableModule)
    if not networks_only:
        check_types += (OptimizerWrapper, DeepSpeedOptimizerWrapper)

    if isinstance(obj, check_types):
        return True
    elif isinstance(obj, Optimizer):
        raise TypeError("Optimizer objects should be wrapped by OptimizerWrapper.")
    if isinstance(obj, dict):
        return any(
            recursive_check_module_attrs(v, networks_only=networks_only)
            for v in obj.values()
        )
    if isinstance(obj, list):
        return any(
            recursive_check_module_attrs(v, networks_only=networks_only) for v in obj
        )
    return False


def chkpt_attribute_to_device(
    chkpt_dict: Dict[str, torch.Tensor], device: str
) -> Dict[str, Any]:
    """Place checkpoint attributes on device. Used when loading saved agents.

    :param chkpt_dict: Checkpoint dictionary
    :type chkpt_dict: dict
    :param device: Device for accelerated computing, 'cpu' or 'cuda'
    :type device: str
    """
    if isinstance(chkpt_dict, list):
        return [chkpt_attribute_to_device(chkpt, device) for chkpt in chkpt_dict]

    assert isinstance(chkpt_dict, dict), f"Expected dict, got {type(chkpt_dict)}"

    for key, value in chkpt_dict.items():
        if isinstance(value, torch.Tensor):
            chkpt_dict[key] = value.to(device)
    return chkpt_dict


def key_in_nested_dict(nested_dict: Dict[str, Any], target: str) -> bool:
    """Helper function to determine if key is in nested dictionary

    :param nested_dict: Nested dictionary
    :type nested_dict: Dict[str, Dict[str, ...]]
    :param target: Target string
    :type target: str
    """
    for k, v in nested_dict.items():
        if k == target:
            return True
        if isinstance(v, dict):
            return key_in_nested_dict(v, target)
    return False


def compile_model(model: Module, mode: Optional[str] = "default") -> Module:
    """Compiles torch model if not already compiled

    :param model: torch model
    :type model: nn.Module
    :param mode: torch compile mode, defaults to "default"
    :type mode: str, optional
    :return: compiled model
    :rtype: OptimizedModule
    """
    return (
        torch.compile(model, mode=mode)
        if not isinstance(model, torch._dynamo.eval_frame.OptimizedModule)
        and mode is not None
        else model
    )


def remove_compile_prefix(state_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Removes _orig_mod prefix on state dict created by torch compile

    :param state_dict: model state dict
    :type state_dict: dict
    :return: state dict with prefix removed
    :rtype: dict
    """
    return OrderedDict(
        [
            (k.split(".", 1)[1], v) if k.startswith("_orig_mod") else (k, v)
            for k, v in state_dict.items()
        ]
    )


SupportedSpace = Union[
    spaces.Box, spaces.Dict, spaces.Tuple, spaces.Discrete, spaces.MultiDiscrete
]


def concatenate_spaces(space_list: List[SupportedSpace]) -> spaces.Space:
    """Concatenates a list of spaces into a single space. If spaces correspond to images,
    we check that their shapes are the same and use the first space's shape as the shape of the
    concatenated space.

    :param spaces: List of spaces to concatenate
    :type spaces: List[spaces.Space]
    :return: Concatenated space
    :rtype: spaces.Space
    """
    if all(isinstance(space, spaces.Dict) for space in space_list):
        return spaces.Dict(
            {
                key: concatenate_spaces([space[key] for space in space_list])
                for key in space_list[0].spaces.keys()
            }
        )

    elif all(isinstance(space, spaces.Tuple) for space in space_list):
        return spaces.Tuple(
            [
                concatenate_spaces([space[i] for space in space_list])
                for i in range(len(space_list[0]))
            ]
        )

    elif all(isinstance(space, spaces.Box) for space in space_list):
        # NOTE: For image spaces the concatenation is handled under-the-hood through the
        # specification of `n_agents` in EvolvableNetwork objects, whereby 3d convolutions
        # are used. This is why we enforce all image spaces to have the same shape.
        if all(is_image_space(space) for space in space_list):
            assert all(space.shape == space_list[0].shape for space in space_list), (
                "AgileRL only supports multi-agent settings with the same shape for the image "
                "spaces of different agents."
            )

            return space_list[0]

        low = np.concatenate([space.low for space in space_list], axis=0)
        high = np.concatenate([space.high for space in space_list], axis=0)
        return spaces.Box(low=low, high=high, dtype=space_list[0].dtype)

    elif all(isinstance(space, spaces.Discrete) for space in space_list):
        n = sum(space.n for space in space_list)
        return spaces.Discrete(n)

    elif all(isinstance(space, spaces.MultiDiscrete) for space in space_list):
        nvec = np.concatenate([space.nvec for space in space_list], axis=0)
        return spaces.MultiDiscrete(nvec)

    else:
        raise TypeError(
            f"Unsupported space types: {set(type(space) for space in spaces)}"
        )


def obs_channels_to_first(observation: NumpyObsType) -> NumpyObsType:
    """Converts observation space from channels last to channels first format.

    :param observation_space: Observation space
    :type observation_space: Union[spaces.Box, spaces.Dict]
    :return: Observation space with channels first format
    :rtype: Union[spaces.Box, spaces.Dict]
    """
    if isinstance(observation, np.ndarray):
        if observation.ndim == 3 or observation.ndim == 4:
            return np.moveaxis(observation, -1, -3)
        else:
            return observation

    elif isinstance(observation, dict):
        return {key: obs_channels_to_first(obs) for key, obs in observation.items()}
    else:
        raise TypeError(f"Expected np.ndarray or dict, got {type(observation)}")


def obs_to_tensor(
    obs: ObservationType, device: Union[str, torch.device]
) -> TorchObsType:
    """
    Moves the observation to the given device as a PyTorch tensor.

    :param obs:
    :type obs: NumpyObsType
    :param device: PyTorch device
    :type device: Union[str, torch.device]
    :return: PyTorch tensor of the observation on a desired device.
    :rtype: TorchObsType
    """
    if isinstance(obs, torch.Tensor):
        return obs.float().to(device)
    if isinstance(obs, np.ndarray):
        return torch.as_tensor(obs, device=device).float()
    elif isinstance(obs, dict):
        return {
            key: torch.as_tensor(_obs, device=device).float()
            for (key, _obs) in obs.items()
        }
    elif isinstance(obs, tuple):
        return tuple(torch.as_tensor(_obs, device=device).float() for _obs in obs)
    elif isinstance(obs, Number):
        return torch.tensor(obs, device=device).float()
    else:
        raise Exception(f"Unrecognized type of observation {type(obs)}")


def maybe_add_batch_dim(
    obs: TorchObsType, space_shape: Tuple[int, ...]
) -> TorchObsType:
    """Adds batch dimension if necessary.

    :param obs: Observation tensor
    :type obs: torch.Tensor[float]
    :param space_shape: Observation space shape
    :type space_shape: Tuple[int, ...]
    :return: Observation tensor with batch dimension
    :rtype: torch.Tensor[float]
    """
    if obs.dim() == len(space_shape):
        obs = obs.unsqueeze(0)
    elif obs.dim() == len(space_shape) + 2:
        obs = obs.view(-1, *space_shape)
    elif obs.dim() != len(space_shape) + 1:
        raise ValueError(
            f"Expected observation to have {len(space_shape) + 1} dimensions, got {obs.dim()}."
        )

    return obs


def preprocess_observation(
    observation: NumpyObsType,
    observation_space: spaces.Space,
    device: Union[str, torch.device] = "cpu",
    normalize_images: bool = True,
) -> TorchObsType:
    """Preprocesses observations for forward pass through neural network.

    :param observations: Observations of environment
    :type observations: ObservationType
    :param observation_space: The observation space of the environment, defaults to the agent's observation space
    :type observation_space: spaces.Space
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to "cpu"
    :type device: Union[str, torch.device], optional
    :param normalize_images: Normalize images from [0. 255] to [0, 1], defaults to True
    :type normalize_images: bool, optional

    :return: Preprocessed observations
    :rtype: torch.Tensor[float] or dict[str, torch.Tensor[float]] or Tuple[torch.Tensor[float], ...]
    """
    observation = obs_to_tensor(observation, device)

    # Preprocess different spaces accordingly
    if isinstance(observation_space, spaces.Dict):
        assert isinstance(observation, dict), f"Expected dict, got {type(observation)}"
        preprocessed_obs = {}
        for key, _obs in observation.items():
            preprocessed_obs[key] = preprocess_observation(
                observation=_obs,
                observation_space=observation_space[key],
                device=device,
                normalize_images=normalize_images,
            )

        return preprocessed_obs

    elif isinstance(observation_space, spaces.Tuple):
        assert isinstance(
            observation, tuple
        ), f"Expected tuple, got {type(observation)}"
        return tuple(
            preprocess_observation(_obs, _space, device, normalize_images)
            for _obs, _space in zip(observation, observation_space.spaces)
        )

    assert isinstance(
        observation, torch.Tensor
    ), f"Expected torch.Tensor, got {type(observation)}"

    if isinstance(observation_space, spaces.Box):
        # Normalize images if applicable and specified
        if len(observation_space.shape) == 3 and normalize_images:
            observation = observation / 255.0

        space_shape = observation_space.shape

    elif isinstance(observation_space, spaces.Discrete):
        # One hot encoding of discrete observation
        observation = F.one_hot(
            observation.long(), num_classes=int(observation_space.n)
        ).float()
        if observation_space.n > 1:
            observation = (
                observation.squeeze()
            )  # If n == 1 then squeeze removes obs dim

        space_shape = (observation_space.n,)

    elif isinstance(observation_space, spaces.MultiDiscrete):
        # Tensor concatenation of one hot encodings of each Categorical sub-space
        observation = torch.cat(
            [
                F.one_hot(
                    obs_.long(), num_classes=int(observation_space.nvec[idx])
                ).float()
                for idx, obs_ in enumerate(torch.split(observation.long(), 1, dim=1))
            ],
            dim=-1,
        )
        space_shape = (sum(observation_space.nvec),)
    else:
        raise TypeError(
            f"AgileRL currently doesn't support {type(observation_space)} spaces."
        )

    # Check add batch dimension if necessary
    observation = maybe_add_batch_dim(observation, space_shape)

    return observation


def apply_image_normalization(
    observation: NumpyObsType, observation_space: spaces.Space
) -> NumpyObsType:
    """Normalize images using minmax scaling

    :param observation: Observation
    :type observation: NumpyObsType
    :param observation_space: Observation space
    :type observation_space: spaces.Space
    :return: Observation
    :rtype: NumpyObsType
    """
    if np.inf in observation_space.high:
        warnings.warn(
            "np.inf detected in observation_space.high, bypassing normalization."
        )
        return observation

    if -np.inf in observation_space.low:
        warnings.warn(
            "-np.inf detected in observation_space.low, bypassing normalization."
        )
        return observation

    if np.all(observation_space.high == 1) and np.all(observation_space.low == 0):
        # minmax scaling
        return observation

    observation = (observation - observation_space.low) / (
        observation_space.high - observation_space.low
    )
    return observation


# TODO: The following functions are currently used in PPO (on-policy) as a means of handling
# experiences in the absence of a rollout buffer -> This will not be needed in the future.
def get_experiences_samples(
    minibatch_indices: np.ndarray, *experiences: TorchObsType
) -> Tuple[TorchObsType, ...]:
    """Samples experiences given minibatch indices.

    :param minibatch_indices: Minibatch indices
    :type minibatch_indices: numpy.ndarray[int]
    :param experiences: Experiences to sample from
    :type experiences: Tuple[torch.Tensor[float], ...]

    :return: Sampled experiences
    :rtype: Tuple[torch.Tensor[float], ...]
    """
    sampled_experiences = []
    for exp in experiences:
        if isinstance(exp, dict):
            sampled_exp = {key: value[minibatch_indices] for key, value in exp.items()}
        elif isinstance(exp, torch.Tensor):
            sampled_exp = exp[minibatch_indices]
        else:
            raise TypeError(f"Unsupported experience type: {type(exp)}")

        sampled_experiences.append(sampled_exp)

    return tuple(sampled_experiences)


def stack_experiences(
    *experiences: MaybeObsList, to_torch: bool = True
) -> Tuple[ArrayOrTensor, ...]:
    """Stacks experiences into a single array or tensor.

    :param experiences: Experiences to stack
    :type experiences: list[numpy.ndarray[float]] or list[dict[str, numpy.ndarray[float]]]
    :param to_torch: If True, convert the stacked experiences to a torch tensor, defaults to True
    :type to_torch: bool, optional

    :return: Stacked experiences
    :rtype: Tuple[ArrayOrTensor, ...]
    """
    stacked_experiences = []
    for exp in experiences:
        # Some cases where an experience just involves e.g. a single "next_state"
        if not isinstance(exp, list):
            stacked_exp = exp
            if to_torch and isinstance(exp, np.ndarray):
                stacked_exp = torch.from_numpy(stacked_exp)

        elif isinstance(exp[0], dict):
            stacked_exp = defaultdict(list)
            for it in exp:
                for key, value in it.items():
                    stacked_exp[key].append(value)

            stacked_exp = {key: np.array(value) for key, value in stacked_exp.items()}
            if to_torch:
                stacked_exp = {
                    key: torch.from_numpy(value) for key, value in stacked_exp.items()
                }

        elif isinstance(exp[0], (np.ndarray, Number)):
            stacked_exp = np.array(exp)
            if to_torch:
                stacked_exp = torch.from_numpy(stacked_exp)

        elif isinstance(exp[0], torch.Tensor):
            stacked_exp = torch.stack(exp)

        else:
            raise TypeError(f"Unsupported experience type: {type(exp[0])}")

        stacked_experiences.append(stacked_exp)

    return tuple(stacked_experiences)


def stack_and_pad_experiences(
    *experiences: MaybeObsList, padding_values: List[Union[int, float, bool]]
) -> Tuple[ArrayOrTensor, ...]:
    """Stacks experiences into a single tensor, padding them to the maximum length.

    :param experiences: Experiences to stack
    :type experiences: list[numpy.ndarray[float]] or list[dict[str, numpy.ndarray[float]]]
    :param to_torch: If True, convert the stacked experiences to a torch tensor, defaults to True
    :type to_torch: bool, optional

    :return: Stacked experiences
    :rtype: Tuple[ArrayOrTensor, ...]
    """
    stacked_experiences = []
    for exp, padding in zip(experiences, padding_values):
        if not isinstance(exp, list):
            stacked_exp = exp
        elif isinstance(exp[0], torch.Tensor):
            max_size = max(e.shape[-1] for e in exp)
            padding_sizes = [(max_size - e.shape[-1]) for e in exp]
            if sum(padding_sizes) != 0:
                exp = [
                    F.pad(e, (0, padding_size), value=padding)
                    for e, padding_size in zip(exp, padding_sizes)
                ]

            stacked_exp = torch.cat(exp, dim=0)
        else:
            raise TypeError(f"Unsupported experience type: {type(exp[0])}")
        stacked_experiences.append(stacked_exp)
    return tuple(stacked_experiences)


def flatten_experiences(*experiences: ArrayOrTensor) -> Tuple[ArrayOrTensor, ...]:
    """Flattens experiences into a single array or tensor.

    :param experiences: Experiences to flatten
    :type experiences: Tuple[numpy.ndarray[float], ...] or Tuple[torch.Tensor[float], ...]

    :return: Flattened experiences
    :rtype: Tuple[numpy.ndarray[float], ...] or Tuple[torch.Tensor[float], ...]
    """

    def flatten(arr: ArrayOrTensor) -> ArrayOrTensor:
        # Need to flatten batch and n_env dimensions
        shape = arr.shape
        if len(shape) < 3:
            shape = (*shape, 1)

        arr = arr.swapaxes(0, 1).reshape(shape[0] * shape[1], *shape[2:])
        return arr

    flattened_experiences = []
    for exp in experiences:
        if isinstance(exp, dict):
            flattened_exp = {key: flatten(value) for key, value in exp.items()}
        elif isinstance(exp, (torch.Tensor, np.ndarray)):
            flattened_exp = flatten(exp)
        else:
            raise TypeError(f"Unsupported experience type: {type(exp)}")

        flattened_experiences.append(flattened_exp)

    return tuple(flattened_experiences)


def is_vectorized_experiences(*experiences: ArrayOrTensor) -> bool:
    """Checks if experiences are vectorised.

    :param experiences: Experiences to check
    :type experiences: Tuple[numpy.ndarray[float], ...] or Tuple[torch.Tensor[float], ...]

    :return: True if experiences are vectorised, False otherwise
    :rtype: bool
    """
    is_vec_ls = []
    for exp in experiences:
        if isinstance(exp, dict):
            is_vec = all(value.ndim > 1 for value in exp.values())
        else:
            is_vec = exp.ndim > 1

        is_vec_ls.append(is_vec)

    return all(is_vec_ls)



from typing import Any, Callable, Dict

import torch
import torch.nn as nn

from agilerl.modules import EvolvableModule
from agilerl.typing import DeviceType


def to_evolvable(
    module_fn: Callable[[], nn.Module],
    module_kwargs: Dict[str, Any],
    device: DeviceType,
) -> EvolvableModule:
    return DummyEvolvable(module_fn, module_kwargs, device)


class DummyEvolvable(EvolvableModule):
    """Wrapper to convert a PyTorch nn.Module into an EvolvableModule object.

    .. note::
        This doesn't actually allow the user to mutate its architecture, but rather allows us
        to use nn.Module objects as networks in an ``EvolvableAlgorithm``. If a user wants to
        mutate the architecture of a network, they should create their network using the ``EvolvableModule``
        class hierarchy directly. Please refer to the documentation for more information on how to do this.

    :param module_fn: Function that returns a PyTorch nn.Module object.
    :type module_fn: Callable[[], nn.Module]
    :param module_kwargs: Keyword arguments to pass to the module_fn.
    :type module_kwargs: Dict[str, Any]
    :param device: Device to run the module on.
    :type device: DeviceType
    """

    def __init__(
        self,
        module_fn: Callable[[], nn.Module],
        module_kwargs: Dict[str, Any],
        device: DeviceType,
    ) -> None:

        # Initialize the module
        module = module_fn(**module_kwargs).to(device)

        super().__init__(device)

        self.module_fn = module_fn
        self.module_kwargs = module_kwargs
        self.module = module

    def change_activation(self, activation: str, output: bool) -> None:
        return

    def forward(self, *args, **kwargs) -> torch.Tensor:
        return self.module(*args, **kwargs)

    def generate(self, *args, **kwargs) -> torch.Tensor:
        if not hasattr(self.module, "generate"):
            raise AttributeError(
                f"Module {self.module_fn} does not have a generate method."
            )
        return self.module.generate(*args, **kwargs)



from dataclasses import asdict
from typing import Optional, Type, Union

import torch
from gymnasium import spaces

from agilerl.modules import EvolvableModule
from agilerl.modules.configs import MlpNetConfig
from agilerl.networks.base import EvolvableNetwork
from agilerl.typing import ConfigType, TorchObsType


class ValueNetwork(EvolvableNetwork):
    """Value functions are used in reinforcement learning to estimate the expected value of a state.
    For any given observation, we predict a single scalar value that represents
    the discounted return from that state. Used in e.g. PPO.

    :param observation_space: Observation space of the environment.
    :type observation_space: spaces.Space
    :param encoder_cls: Encoder class to use for the network. Defaults to None, whereby it is
        automatically built using an AgileRL module according the observation space.
    :type encoder_cls: Optional[Union[str, Type[EvolvableModule]]]
    :param encoder_config: Configuration of the encoder.
    :type encoder_config: ConfigType
    :param head_config: Configuration of the head.
    :type head_config: Optional[ConfigType]
    :param min_latent_dim: Minimum latent dimension.
    :type min_latent_dim: int
    :param max_latent_dim: Maximum latent dimension.
    :type max_latent_dim: int
    :param n_agents: Number of agents.
    :type n_agents: Optional[int]
    :param latent_dim: Latent dimension.
    :type latent_dim: int
    :param device: Device to run the network on.
    :type device: str
    """

    def __init__(
        self,
        observation_space: spaces.Space,
        encoder_cls: Optional[Union[str, Type[EvolvableModule]]] = None,
        encoder_config: Optional[ConfigType] = None,
        head_config: Optional[ConfigType] = None,
        min_latent_dim: int = 8,
        max_latent_dim: int = 128,
        n_agents: Optional[int] = None,
        latent_dim: int = 32,
        simba: bool = False,
        device: str = "cpu",
    ):

        super().__init__(
            observation_space,
            encoder_cls=encoder_cls,
            encoder_config=encoder_config,
            action_space=None,
            min_latent_dim=min_latent_dim,
            max_latent_dim=max_latent_dim,
            n_agents=n_agents,
            latent_dim=latent_dim,
            simba=simba,
            device=device,
        )

        if head_config is None:
            head_config = asdict(MlpNetConfig(hidden_size=[16], output_activation=None))

        # Build the network head
        self.build_network_head(head_config)

    def get_output_dense(self) -> torch.nn.Linear:
        """Returns the output dense layer of the network.

        :return: Output dense layer.
        :rtype: torch.nn.Linear
        """
        return self.head_net.get_output_dense()

    def build_network_head(self, net_config: Optional[ConfigType] = None) -> None:
        """Builds the head of the network.

        :param net_config: Configuration of the head.
        :type net_config: Optional[ConfigType]
        """
        self.head_net = self.create_mlp(
            num_inputs=self.latent_dim,
            num_outputs=1,
            name="value",
            net_config=net_config,
        )

    def forward(self, x: TorchObsType) -> torch.Tensor:
        """Forward pass of the network.

        :param x: Input tensor.
        :type x: torch.Tensor, dict[str, torch.Tensor], or list[torch.Tensor]
        :return: Output tensor.
        :rtype: torch.Tensor
        """
        return self.head_net(self.encoder(x))

    def recreate_network(self) -> None:
        """Recreates the network."""
        self.recreate_encoder()

        head_net = self.create_mlp(
            num_inputs=self.latent_dim,
            num_outputs=1,
            name="value",
            net_config=self.head_net.net_config,
        )

        self.head_net = EvolvableModule.preserve_parameters(self.head_net, head_net)



from typing import Any, Dict, Optional, Union

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from gymnasium import spaces

from agilerl.algorithms.core import RLAlgorithm
from agilerl.algorithms.core.registry import HyperparameterConfig, NetworkGroup
from agilerl.algorithms.core.wrappers import OptimizerWrapper
from agilerl.modules.base import EvolvableModule
from agilerl.modules.cnn import EvolvableCNN
from agilerl.modules.mlp import EvolvableMLP
from agilerl.modules.multi_input import EvolvableMultiInput
from agilerl.networks.base import EvolvableNetwork
from agilerl.networks.value_networks import ValueNetwork
from agilerl.typing import ArrayLike, ExperiencesType, GymEnvType, ObservationType
from agilerl.utils.algo_utils import make_safe_deepcopies, obs_channels_to_first
from agilerl.utils.evolvable_networks import get_default_encoder_config

SupportedEvolvable = Union[
    EvolvableCNN, EvolvableMLP, EvolvableMultiInput, EvolvableNetwork
]


class NeuralTS(RLAlgorithm):
    """The NeuralTS algorithm class. NeuralTS paper: https://arxiv.org/abs/2010.00827

    :param observation_space: Observation space of the environment
    :type observation_space: gym.spaces.Space
    :param action_space: Action space of the environment
    :type action_space: gym.spaces.Space
    :param index: Index to keep track of object instance during tournament selection and mutation, defaults to 0
    :type index: int, optional
    :param hp_config: RL hyperparameter mutation configuration, defaults to None, whereby algorithm mutations are disabled.
    :type hp_config: HyperparameterConfig, optional
    :param net_config: Network configuration, defaults to None
    :type net_config: dict, optional
    :param gamma: Positive scaling factor, defaults to 1.0
    :type gamma: float, optional
    :param lamb: Regularization parameter lambda, defaults to 1.0
    :type lamb: float, optional
    :param reg: Loss regularization parameter, defaults to 0.000625
    :type reg: float, optional
    :param batch_size: Size of batched sample from replay buffer for learning, defaults to 64
    :type batch_size: int, optional
    :param lr: Learning rate for optimizer, defaults to 3e-3
    :type lr: float, optional
    :param normalize_images: Normalize images flag, defaults to True
    :type normalize_images: bool, optional
    :param learn_step: Learning frequency, defaults to 2
    :type learn_step: int, optional
    :param mut: Most recent mutation to agent, defaults to None
    :type mut: str, optional
    :param actor_network: Custom actor network, defaults to None
    :type actor_network: EvolvableModule, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wrap: Wrap models for distributed training upon creation, defaults to True
    :type wrap: bool, optional
    """

    def __init__(
        self,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        index: int = 0,
        hp_config: Optional[HyperparameterConfig] = None,
        net_config: Optional[Dict[str, Any]] = None,
        gamma: float = 1.0,
        lamb: float = 1.0,
        reg: float = 0.000625,
        batch_size: int = 64,
        lr: float = 3e-3,
        normalize_images: bool = True,
        learn_step: int = 2,
        mut: Optional[str] = None,
        actor_network: Optional[EvolvableModule] = None,
        device: str = "cpu",
        accelerator: Optional[Any] = None,
        wrap: bool = True,
    ):
        super().__init__(
            observation_space,
            action_space,
            index=index,
            hp_config=hp_config,
            device=device,
            accelerator=accelerator,
            normalize_images=normalize_images,
            name="NeuralTS",
        )

        assert learn_step >= 1, "Learn step must be greater than or equal to one."
        assert isinstance(learn_step, int), "Learn step rate must be an integer."
        assert isinstance(
            gamma, (float, int)
        ), "Scaling factor must be a float or integer."
        assert gamma > 0, "Scaling factor must be positive."
        assert isinstance(
            lamb, (float, int)
        ), "Regularization parameter lambda must be a float or integer."
        assert lamb > 0, "Regularization parameter lambda must be greater than zero."
        assert isinstance(reg, float), "Loss regularization parameter must be a float."
        assert reg > 0, "Loss regularization parameter must be greater than zero."
        assert isinstance(batch_size, int), "Batch size must be an integer."
        assert batch_size >= 1, "Batch size must be greater than or equal to one."
        assert isinstance(lr, float), "Learning rate must be a float."
        assert lr > 0, "Learning rate must be greater than zero."
        assert isinstance(
            wrap, bool
        ), "Wrap models flag must be boolean value True or False."

        self.gamma = gamma
        self.learn_step = learn_step
        self.lamb = lamb
        self.reg = reg
        self.batch_size = batch_size
        self.lr = lr
        self.mut = mut
        self.net_config = net_config
        self.regret = [0]

        if actor_network is not None:
            if not isinstance(actor_network, EvolvableModule):
                raise TypeError(
                    f"'actor_network' argument is of type {type(actor_network)}, but must be of type EvolvableModule."
                )

            # Need to make deepcopies for target and detached networks
            self.actor = make_safe_deepcopies(actor_network)
        else:
            net_config = {} if net_config is None else net_config
            simba = net_config.get("simba", False)
            encoder_config = (
                get_default_encoder_config(observation_space, simba)
                if net_config.get("encoder_config") is None
                else net_config["encoder_config"]
            )

            if not simba:
                encoder_config["layer_norm"] = (
                    False  # Layer norm is not used in the original implementation
                )

            net_config["encoder_config"] = encoder_config

            self.actor = ValueNetwork(
                observation_space=observation_space, device=self.device, **net_config
            )

        self.optimizer = OptimizerWrapper(optim.Adam, networks=self.actor, lr=self.lr)

        if self.accelerator is not None and wrap:
            self.wrap_models()

        # Initialize network layers
        self.actor.init_weights_gaussian(std_coeff=4, output_coeff=2)
        self.init_params()

        self.criterion = nn.MSELoss()

        # Register network groups for mutations
        self.register_init_hook(self.init_params)
        self.register_network_group(
            NetworkGroup(eval=self.actor, shared=None, policy=True)
        )

    def init_params(self) -> None:
        """Initializes parameters for the agent network."""
        self.exp_layer = self.actor.get_output_dense()

        self.numel = sum(
            w.numel() for w in self.exp_layer.parameters() if w.requires_grad
        )
        self.sigma_inv = self.lamb * torch.eye(self.numel).to(self.device)
        self.theta_0 = torch.cat(
            [w.flatten() for w in self.exp_layer.parameters() if w.requires_grad]
        )

    def get_action(
        self, obs: ObservationType, action_mask: Optional[ArrayLike] = None
    ) -> int:
        """Returns the next action to take in the environment.

        :param obs: State observation, or multiple observations in a batch
        :type obs: numpy.ndarray[float]
        :param action_mask: Mask of legal actions 1=legal 0=illegal, defaults to None
        :type action_mask: numpy.ndarray, optional
        :return: Action to take in the environment
        :rtype: int
        """
        obs = self.preprocess_observation(obs)

        mu = self.actor(obs)
        g = torch.zeros((self.action_dim, self.numel)).to(self.device)
        for k, fx in enumerate(mu):
            self.optimizer.zero_grad()
            fx.backward(retain_graph=True)
            g[k] = torch.cat(
                [
                    w.grad.detach().flatten() / np.sqrt(self.exp_layer.weight.size(0))
                    for w in self.exp_layer.parameters()
                    if w.requires_grad
                ]
            )

        with torch.no_grad():
            action_values = torch.normal(
                mean=self.actor(obs),
                std=self.gamma
                * torch.sqrt(
                    torch.matmul(
                        torch.matmul(g[:, None, :], self.sigma_inv), g[:, :, None]
                    )[:, 0, :]
                ),
            )

        action_values = action_values.cpu().numpy()
        if action_mask is None:
            action = np.argmax(action_values)
        else:
            inv_mask = 1 - action_mask
            masked_action_values = np.ma.array(action_values, mask=inv_mask)
            action = np.argmax(masked_action_values)

        # Sherman-Morrison-Woodbury Update
        v = g[action].unsqueeze(-1)
        self.sigma_inv -= (self.sigma_inv @ v @ v.T @ self.sigma_inv) / (
            1 + v.T @ self.sigma_inv @ v
        )

        return action

    def learn(self, experiences: ExperiencesType) -> float:
        """Updates agent network parameters to learn from experiences.

        :param experiences: Batched states, rewards in that order.
        :type obs: list[torch.Tensor[float]]
        """
        states, rewards = experiences

        states = states.to(self.device)
        rewards = rewards.to(self.device)

        pred_rewards = self.actor(states)

        # loss backprop
        loss = self.criterion(rewards, pred_rewards)
        loss += (
            self.reg
            * torch.norm(
                torch.cat(
                    [
                        w.flatten()
                        for w in self.exp_layer.parameters()
                        if w.requires_grad
                    ]
                )
                - self.theta_0
            )
            ** 2
        )
        self.optimizer.zero_grad()
        if self.accelerator is not None:
            self.accelerator.backward(loss)
        else:
            loss.backward()

        self.optimizer.step()

        return loss.item()

    def test(
        self,
        env: GymEnvType,
        swap_channels: bool = False,
        max_steps: int = 100,
        loop: int = 1,
    ) -> float:
        """Returns mean test score of agent in environment with epsilon-greedy policy.

        :param env: The environment to be tested in
        :type env: Gym-style environment
        :param swap_channels: Swap image channels dimension from last to first [H, W, C] -> [C, H, W], defaults to False
        :type swap_channels: bool, optional
        :param max_steps: Maximum number of testing steps, defaults to 500
        :type max_steps: int, optional
        :param loop: Number of testing loops/episodes to complete. The returned score is the mean over these tests. Defaults to 3
        :type loop: int, optional
        """
        self.set_training_mode(False)
        with torch.no_grad():
            rewards = []
            for i in range(loop):
                obs = env.reset()
                score = 0
                for _ in range(max_steps):
                    if swap_channels:
                        obs = obs_channels_to_first(obs)
                    obs = torch.from_numpy(obs).float()
                    obs = obs.to(self.device)
                    action = np.argmax(self.actor(obs).cpu().numpy())
                    obs, reward = env.step(action)
                    score += reward
                rewards.append(score)
        mean_fit = np.mean(rewards)
        self.fitness.append(mean_fit)
        return mean_fit



from typing import Any, Dict, Optional, Tuple

import numpy as np
import torch
import torch.optim as optim
from gymnasium import spaces
from torch.nn.utils import clip_grad_norm_

from agilerl.algorithms.core import RLAlgorithm
from agilerl.algorithms.core.registry import HyperparameterConfig, NetworkGroup
from agilerl.algorithms.core.wrappers import OptimizerWrapper
from agilerl.modules.base import EvolvableModule
from agilerl.modules.configs import MlpNetConfig
from agilerl.networks.q_networks import RainbowQNetwork
from agilerl.typing import (
    ArrayLike,
    ExperiencesType,
    GymEnvType,
    ObservationType,
    TorchObsType,
)
from agilerl.utils.algo_utils import make_safe_deepcopies, obs_channels_to_first
from agilerl.wrappers.make_evolvable import MakeEvolvable


class RainbowDQN(RLAlgorithm):
    """The Rainbow DQN algorithm class. Rainbow DQN paper: https://arxiv.org/abs/1710.02298

    :param observation_space: Observation space of the environment
    :type observation_space: gym.spaces.Space
    :param action_space: Action space of the environment
    :type action_space: gym.spaces.Space
    :param index: Index to keep track of object instance during tournament selection and mutation, defaults to 0
    :type index: int, optional
    :param hp_config: RL hyperparameter mutation configuration, defaults to None, whereby algorithm mutations are disabled.
    :type hp_config: HyperparameterConfig, optional
    :param net_config: Network configuration, defaults to None
    :type net_config: dict, optional
    :param batch_size: Size of batched sample from replay buffer for learning, defaults to 64
    :type batch_size: int, optional
    :param lr: Learning rate for optimizer, defaults to 1e-4
    :type lr: float, optional
    :param learn_step: Learning frequency, defaults to 5
    :type learn_step: int, optional
    :param gamma: Discount factor, defaults to 0.99
    :type gamma: float, optional
    :param tau: For soft update of target network parameters, defaults to 1e-3
    :type tau: float, optional
    :param beta: Importance sampling coefficient, defaults to 0.4
    :type beta: float, optional
    :param prior_eps: Minimum priority for sampling, defaults to 1e-6
    :type prior_eps: float, optional
    :param num_atoms: Unit number of support, defaults to 51
    :type num_atoms: int, optional
    :param v_min: Minimum value of support, defaults to 0
    :type v_min: float, optional
    :param v_max: Maximum value of support, defaults to 200
    :type v_max: float, optional
    :param noise_std: Noise standard deviation, defaults to 0.5
    :type noise_std: float, optional
    :param n_step: Step number to calculate n-step td error, defaults to 3
    :type n_step: int, optional
    :param mut: Most recent mutation to agent, defaults to None
    :type mut: str, optional
    :param combined_reward: Boolean flag indicating whether to use combined 1-step and n-step reward, defaults to False
    :type combined_reward: bool, optional
    :param actor_network: Custom actor network, defaults to None
    :type actor_network: nn.Module, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wrap: Wrap models for distributed training upon creation, defaults to True
    :type wrap: bool, optional
    """

    def __init__(
        self,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        index: int = 0,
        hp_config: Optional[HyperparameterConfig] = None,
        net_config: Optional[Dict[str, Any]] = None,
        batch_size: int = 64,
        lr: float = 1e-4,
        learn_step: int = 5,
        gamma: float = 0.99,
        tau: float = 1e-3,
        beta: float = 0.4,
        prior_eps: float = 1e-6,
        num_atoms: int = 51,
        v_min: float = 0,
        v_max: float = 200,
        noise_std: float = 0.5,
        n_step: int = 3,
        mut: Optional[str] = None,
        normalize_images: bool = True,
        combined_reward: bool = False,
        actor_network: Optional[EvolvableModule] = None,
        device: str = "cpu",
        accelerator: Optional[Any] = None,
        wrap: bool = True,
    ) -> None:
        super().__init__(
            observation_space,
            action_space,
            index=index,
            hp_config=hp_config,
            device=device,
            accelerator=accelerator,
            normalize_images=normalize_images,
            name="Rainbow DQN",
        )

        assert learn_step >= 1, "Learn step must be greater than or equal to one."
        assert isinstance(learn_step, int), "Learn step rate must be an integer."
        assert isinstance(batch_size, int), "Batch size must be an integer."
        assert batch_size >= 1, "Batch size must be greater than or equal to one."
        assert isinstance(lr, float), "Learning rate must be a float."
        assert lr > 0, "Learning rate must be greater than zero."
        assert isinstance(learn_step, int), "Learn step rate must be an integer."
        assert learn_step >= 1, "Learn step must be greater than or equal to one."
        assert isinstance(gamma, (float, int, torch.Tensor)), "Gamma must be a float."
        assert isinstance(tau, float), "Tau must be a float."
        assert tau > 0, "Tau must be greater than zero."
        assert isinstance(
            prior_eps, float
        ), "Minimum priority for sampling must be a float."
        assert prior_eps > 0, "Minimum priority for sampling must be greater than zero."
        assert isinstance(num_atoms, int), "Number of atoms must be an integer."
        assert num_atoms >= 1, "Number of atoms must be greater than or equal to one."
        assert isinstance(
            v_min, (float, int)
        ), "Minimum value of support must be a float."
        assert isinstance(
            v_max, (float, int)
        ), "Maximum value of support must be a float."
        assert (
            v_max >= v_min
        ), "Maximum value of support must be greater than or equal to minimum value."
        assert isinstance(n_step, int), "Step number must be an integer."
        assert n_step >= 1, "Step number must be greater than or equal to one."
        assert isinstance(
            wrap, bool
        ), "Wrap models flag must be boolean value True or False."

        self.batch_size = batch_size
        self.learn_step = learn_step
        self.lr = lr
        self.gamma = gamma
        self.tau = tau
        self.beta = beta
        self.prior_eps = prior_eps
        self.num_atoms = num_atoms
        self.net_config = net_config
        self.v_min = v_min
        self.v_max = v_max
        self.n_step = n_step
        self.mut = mut
        self.combined_reward = combined_reward
        self.noise_std = noise_std

        self.support = torch.linspace(
            self.v_min, self.v_max, self.num_atoms, device=self.device
        )
        self.delta_z = (self.v_max - self.v_min) / (self.num_atoms - 1)

        if actor_network is not None:
            if isinstance(actor_network, MakeEvolvable):
                actor_network.rainbow = True
                actor_network = actor_network
                actor_network.support = self.support
                actor_network.num_atoms = self.num_atoms
                actor_network = MakeEvolvable(**actor_network.init_dict)
                actor_network.load_state_dict(actor_network.state_dict())
            elif not isinstance(actor_network, EvolvableModule):
                raise TypeError(
                    f"'actor_network' argument is of type {type(actor_network)}, but must be of type EvolvableModule."
                )

            self.actor, self.actor_target = make_safe_deepcopies(
                actor_network, actor_network
            )
        else:
            net_config = {} if net_config is None else net_config
            head_config: Optional[Dict[str, Any]] = net_config.get("head_config", None)

            head_config = MlpNetConfig(
                hidden_size=(
                    [64]
                    if head_config is None
                    else head_config.get("hidden_size", [64])
                ),
                noise_std=self.noise_std,
                output_activation="ReLU",
            )
            net_config["head_config"] = head_config

            def create_actor():
                return RainbowQNetwork(
                    observation_space=observation_space,
                    action_space=action_space,
                    support=self.support,
                    num_atoms=self.num_atoms,
                    noise_std=self.noise_std,
                    device=self.device,
                    **net_config,
                )

            self.actor = create_actor()
            self.actor_target = create_actor()

        # Create the target network by copying the actor network
        self.actor_target.load_state_dict(self.actor.state_dict())

        # Optimizer
        self.optimizer = OptimizerWrapper(optim.Adam, networks=self.actor, lr=self.lr)

        if self.accelerator is not None and wrap:
            self.wrap_models()

        # Put the nets into training mode
        self.actor.train()
        self.actor_target.train()

        # Register network groups for mutations
        self.register_network_group(
            NetworkGroup(eval=self.actor, shared=self.actor_target, policy=True)
        )

    def get_action(
        self,
        obs: ObservationType,
        action_mask: Optional[np.ndarray] = None,
        training: bool = True,
    ) -> ArrayLike:
        """Returns the next action to take in the environment.

        :param obs: State observation, or multiple observations in a batch
        :type obs: numpy.ndarray[float]
        :param action_mask: Mask of legal actions 1=legal 0=illegal, defaults to None
        :type action_mask: numpy.ndarray, optional
        :param training: Flag indicating whether the model is in training mode, defaults to True
        :type training: bool, optional
        :return: The action to take
        :rtype: numpy.ndarray
        """
        obs = self.preprocess_observation(obs)

        self.actor.train(mode=training)
        with torch.no_grad():
            action_values = self.actor(obs)

        if action_mask is None:
            action = np.argmax(action_values.cpu().data.numpy(), axis=-1)
        else:
            # Need to stack if vectorized env
            action_mask = (
                np.stack(action_mask)
                if action_mask.dtype == np.object_ or isinstance(action_mask, list)
                else action_mask
            )
            inv_mask = 1 - action_mask
            masked_action_values = np.ma.array(
                action_values.cpu().data.numpy(), mask=inv_mask
            )
            action = np.argmax(masked_action_values, axis=-1)

        self.actor.train()

        return action

    def _dqn_loss(
        self,
        states: TorchObsType,
        actions: torch.Tensor,
        rewards: torch.Tensor,
        next_states: torch.Tensor,
        dones: torch.Tensor,
        gamma: float,
    ) -> torch.Tensor:
        """Calculates the DQN loss.

        :param states: Batch of current states
        :type states: torch.Tensor
        :param actions: Batch of actions taken
        :type actions: torch.Tensor
        :param rewards: Batch of rewards received
        :type rewards: torch.Tensor
        :param next_states: Batch of next states
        :type next_states: torch.Tensor
        :param dones: Batch of done flags indicating episode termination
        :type dones: torch.Tensor
        :param gamma: Discount factor
        :type gamma: float
        :return: Element-wise loss
        :rtype: torch.Tensor
        """
        states = self.preprocess_observation(states)
        next_states = self.preprocess_observation(next_states)

        with torch.no_grad():

            # Predict next actions from next_states
            next_actions = self.actor(next_states).argmax(1)

            # Predict the target q distribution for the same next states
            target_q_dist = self.actor_target(next_states, q=False)

            # Index the target q_dist to select the distributions corresponding to next_actions
            target_q_dist = target_q_dist[range(self.batch_size), next_actions]

            # Determine the target z values
            t_z = rewards + (1 - dones) * gamma * self.support
            t_z = t_z.clamp(min=self.v_min, max=self.v_max)

            # Finds closest support element index value
            b = (t_z - self.v_min) / self.delta_z

            # Find the neighbouring indices of b
            L = b.floor().long()
            u = b.ceil().long()

            # Shape of projected q distribution is (batch_size, num_atoms) as we have argmaxed over actions
            # Fix disappearing probability mass
            L[(u > 0) * (L == u)] -= 1
            u[(L < (self.num_atoms - 1)) * (L == u)] += 1
            offset = (
                torch.linspace(
                    0,
                    (self.batch_size - 1) * self.num_atoms,
                    self.batch_size,
                    device=self.device,
                )
                .long()
                .unsqueeze(1)
                .expand(self.batch_size, self.num_atoms)
            )
            proj_dist = torch.zeros(target_q_dist.size(), device=self.device)

            proj_dist.view(-1).index_add_(
                0, (L + offset).view(-1), (target_q_dist * (u.float() - b)).view(-1)
            )
            proj_dist.view(-1).index_add_(
                0, (u + offset).view(-1), (target_q_dist * (b - L.float())).view(-1)
            )

        # Calculate the current obs
        log_q_dist = self.actor(states, q=False, log=True)
        log_p = log_q_dist[range(self.batch_size), actions.squeeze().long()]

        # loss
        elementwise_loss = -(proj_dist * log_p).sum(1)
        return elementwise_loss

    def learn(
        self, experiences: ExperiencesType, n_step: bool = False, per: bool = False
    ) -> Tuple[float, Optional[ArrayLike], Optional[ArrayLike]]:
        """Updates agent network parameters to learn from experiences.

        :param experiences: List of batched states, actions, rewards, next_states, dones in that order.
        :type obs: list[torch.Tensor[float]]
        :param n_step: Use multi-step learning, defaults to True
        :type n_step: bool, optional
        :param per: Use prioritized experience replay buffer, defaults to True
        :type per: bool, optional

        :return: Tuple of loss, indices, and new priorities
        :rtype: Tuple[float, numpy.ndarray, numpy.ndarray]
        """
        experiences = self.to_device(*experiences)
        if per:
            if n_step:
                (
                    states,
                    actions,
                    rewards,
                    next_states,
                    dones,
                    weights,
                    idxs,
                    n_states,
                    n_actions,
                    n_rewards,
                    n_next_states,
                    n_dones,
                ) = experiences
            else:
                (
                    states,
                    actions,
                    rewards,
                    next_states,
                    dones,
                    weights,
                    idxs,
                ) = experiences

            if self.combined_reward or not n_step:
                elementwise_loss = self._dqn_loss(
                    states, actions, rewards, next_states, dones, self.gamma
                )
            if n_step:
                n_gamma = self.gamma**self.n_step
                n_step_elementwise_loss = self._dqn_loss(
                    n_states, n_actions, n_rewards, n_next_states, n_dones, n_gamma
                )
                if self.combined_reward:
                    elementwise_loss += n_step_elementwise_loss
                else:
                    elementwise_loss = n_step_elementwise_loss
            loss = torch.mean(elementwise_loss * weights)

        else:
            if n_step:
                (
                    states,
                    actions,
                    rewards,
                    next_states,
                    dones,
                    idxs,
                    n_states,
                    n_actions,
                    n_rewards,
                    n_next_states,
                    n_dones,
                ) = experiences
            else:
                (
                    states,
                    actions,
                    rewards,
                    next_states,
                    dones,
                ) = experiences

            if not n_step:
                idxs = None

            new_priorities = None

            if self.combined_reward or not n_step:
                elementwise_loss = self._dqn_loss(
                    states, actions, rewards, next_states, dones, self.gamma
                )

            if n_step:
                n_gamma = self.gamma**self.n_step
                n_step_elementwise_loss = self._dqn_loss(
                    n_states, n_actions, n_rewards, n_next_states, n_dones, n_gamma
                )
                if self.combined_reward:
                    elementwise_loss += n_step_elementwise_loss
                else:
                    elementwise_loss = n_step_elementwise_loss
            loss = torch.mean(elementwise_loss)

        self.optimizer.zero_grad()
        if self.accelerator is not None:
            self.accelerator.backward(loss)
        else:
            loss.backward()

        clip_grad_norm_(self.actor.parameters(), 10.0)
        self.optimizer.step()

        # soft update target network
        self.soft_update()
        self.actor.reset_noise()
        self.actor_target.reset_noise()

        if per:
            loss_for_prior = elementwise_loss.detach().cpu().numpy()
            new_priorities = loss_for_prior + self.prior_eps

        return loss.item(), idxs, new_priorities

    def soft_update(self) -> None:
        """Soft updates target network."""
        for eval_param, target_param in zip(
            self.actor.parameters(), self.actor_target.parameters()
        ):
            target_param.data.copy_(
                self.tau * eval_param.data + (1.0 - self.tau) * target_param.data
            )

    def test(
        self,
        env: GymEnvType,
        swap_channels: bool = False,
        max_steps: Optional[int] = None,
        loop: int = 3,
    ) -> float:
        """Returns mean test score of agent in environment with epsilon-greedy policy.

        :param env: The environment to be tested in
        :type env: Gym-style environment
        :param swap_channels: Swap image channels dimension from last to first [H, W, C] -> [C, H, W], defaults to False
        :type swap_channels: bool, optional
        :param max_steps: Maximum number of testing steps, defaults to None
        :type max_steps: int, optional
        :param loop: Number of testing loops/episodes to complete. The returned score is the mean over these tests. Defaults to 3
        :type loop: int, optional
        """
        self.set_training_mode(False)
        with torch.no_grad():
            rewards = []
            num_envs = env.num_envs if hasattr(env, "num_envs") else 1
            for _ in range(loop):
                obs, info = env.reset()
                scores = np.zeros(num_envs)
                completed_episode_scores = np.zeros(num_envs)
                finished = np.zeros(num_envs)
                step = 0
                while not np.all(finished):
                    if swap_channels:
                        obs = obs_channels_to_first(obs)

                    action_mask = info.get("action_mask", None)
                    action = self.get_action(
                        obs, training=False, action_mask=action_mask
                    )
                    obs, reward, done, trunc, info = env.step(action)
                    step += 1
                    scores += np.array(reward)
                    for idx, (d, t) in enumerate(zip(done, trunc)):
                        if (
                            d or t or (max_steps is not None and step == max_steps)
                        ) and not finished[idx]:
                            completed_episode_scores[idx] = scores[idx]
                            finished[idx] = 1
                rewards.append(np.mean(completed_episode_scores))
        mean_fit = np.mean(rewards)
        self.fitness.append(mean_fit)
        return mean_fit



import copy
import inspect
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np
import torch
import torch.nn as nn
from numpy.random import Generator

from agilerl.modules.custom_components import NoisyLinear
from agilerl.protocols import MutationMethod, MutationType
from agilerl.utils.algo_utils import recursive_check_module_attrs

SelfEvolvableModule = TypeVar("SelfEvolvableModule", bound="EvolvableModule")


def is_evolvable(attr: str, obj: Any) -> bool:
    """Check if an attribute of a module is evolvable.

    :param attr: The attribute name.
    :type attr: str
    :param obj: The attribute object.
    :type obj: Any
    :return: True if the attribute is evolvable, False otherwise.
    :rtype: bool
    """
    return (
        recursive_check_module_attrs(obj, networks_only=True)
        and not attr.startswith("_")
        and not attr.endswith("_")
    )


def mutation(
    mutation_type: MutationType, **recreate_kwargs
) -> Callable[[Callable], MutationMethod]:
    """Decorator to register a method as a mutation function of a specific type.

    :param mutation_type: The type of mutation function.
    :type mutation_type: MutationType
    :return: The decorator function.
    :rtype: Callable[[Callable], MutationMethod]
    """

    def decorator(func: Callable[[Any], Optional[Dict[str, Any]]]) -> MutationMethod:
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        # Explicitly set the mutation type attribute on the wrapper function
        wrapper._mutation_type = mutation_type

        # Each mutation method might want to call different arguments when recreating the network
        wrapper._recreate_kwargs = recreate_kwargs
        return wrapper

    return decorator


class MutationContext:
    """Tracks nested mutation method calls. This allows us to automatically call a modules
    `recreate_network` method after the outermost mutation method has been called. It handles
    cases where we fall back on another mutation method when a limit has been reached e.g. in
    the EvolvableMLP `add_layer` method, after the maximum number of layers has been reached we
    instead call `add_node`. In cases like this, we want to avoid redundant recreations of the
    network.

    :param module: The evolvable module.
    :type module: EvolvableModule
    :param method: The mutation method.
    :type method: MutationMethod
    :param attribute: The mutation attribute.
    :type attribute: str
    """

    def __init__(
        self, module: SelfEvolvableModule, method: MutationMethod, attribute: str
    ):
        self.module = module
        self.method = method
        self.method_name = attribute

    def __enter__(self):
        self.module._mutation_depth += 1
        self.module.last_mutation = self.method
        self.module.last_mutation_attr = self.method_name
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.module._mutation_depth -= 1

        if self.module._mutation_depth == 0:
            # Check recursively for nested mutations to update last_mutation_attr
            final_mutation_attr = self._resolve_final_mutation_attr()
            self.module.last_mutation_attr = final_mutation_attr

            if final_mutation_attr is not None:
                self.module.last_mutation = getattr(self.module, final_mutation_attr)

                if "." not in final_mutation_attr and not isinstance(
                    self.module, EvolvableWrapper
                ):
                    if self.method._recreate_kwargs:
                        recreation_kwargs = inspect.signature(
                            self.module.recreate_network
                        ).parameters
                        rec_kwargs = {
                            k: v
                            for k, v in self.method._recreate_kwargs.items()
                            if k in recreation_kwargs
                        }
                    else:
                        rec_kwargs = {}

                    self.module.recreate_network(**rec_kwargs)

            if self.module._mutation_hook is not None:
                self.module._mutation_hook()

    def _resolve_final_mutation_attr(self) -> str:
        """Recursively resolve the final mutation attribute."""
        if (
            self.module.last_mutation_attr is not None
            and "." in self.module.last_mutation_attr
        ):
            parts = self.module.last_mutation_attr.split(".")
            mutated_module = self.module
            for part in parts[:-1]:
                mutated_module: SelfEvolvableModule = getattr(mutated_module, part)

            if mutated_module.last_mutation_attr is None:
                return

            return ".".join(parts[:-1] + [mutated_module.last_mutation_attr])

        elif isinstance(self.module, EvolvableWrapper):
            return self.module.wrapped.last_mutation_attr

        return self.module.last_mutation_attr


def _mutation_wrapper(
    module: SelfEvolvableModule, method: MutationMethod, attribute: str
) -> Callable:
    """Wraps mutation methods to use context manager.

    :param module: The evolvable module.
    :type module: EvolvableModule
    :param method: The mutation method.
    :type method: MutationMethod
    :param attribute: The mutation attribute.
    :type attribute: str

    :return: The wrapped mutation method.
    :rtype: Callable
    """

    @wraps(method)
    def wrapped(*args, **kwargs):
        with MutationContext(module, method, attribute):
            if attribute not in module.mutation_methods:
                module.last_mutation_attr = None
                module.last_mutation = None
                return

            return method(*args, **kwargs)

    return wrapped


def _get_filtered_methods(
    module: SelfEvolvableModule, mut_type: MutationType
) -> List[str]:
    """Gets the mutation methods of a given type for the module.

    :param module: The evolvable module.
    :type module: EvolvableModule
    :param mut_type: The type of mutation method.
    :type mut_type: MutationType
    :return: The filtered mutation methods.
    :rtype: List[str]
    """
    _fetch = (
        "layer_mutation_methods"
        if mut_type == MutationType.LAYER
        else "node_mutation_methods"
    )

    init_methods: List[str] = getattr(module, "_" + _fetch)
    _filtered = []
    for method in init_methods:
        if "." not in method:
            _filtered.append(method)
            continue

        comps = method.split(".")
        mod: EvolvableModule = getattr(module, comps[0])
        layer_muts = getattr(mod, _fetch)
        if any(".".join([comps[0], m]) == method for m in layer_muts):
            _filtered.append(method)

    return _filtered


# TODO: Think of a way that doesn't require the use of a metaclass
class ModuleMeta(type):
    """Metaclass to parse the mutation methods of an EvolvableModule instance
    and its superclasses."""

    def __call__(cls, *args, **kwargs):
        instance: SelfEvolvableModule = super().__call__(*args, **kwargs)

        # Wrap mutation methods to use context manager
        for name, method in instance.get_mutation_methods().items():
            setattr(instance, name, _mutation_wrapper(instance, method, name))

        return instance


class EvolvableModule(nn.Module, metaclass=ModuleMeta):
    """Base class for evolvable neural networks.

    :param device: The device to run the network on.
    :type device: str
    """

    def __init__(self, device: str) -> None:
        nn.Module.__init__(self)
        self._init_surface_methods()
        self.device = device

        self._last_mutation = None
        self._last_mutation_attr = None
        self._mutation_hook = None
        self._mutation_depth = 0

    @property
    def init_dict(self) -> Dict[str, Any]:
        return self.get_init_dict()

    @property
    def net_config(self) -> Dict[str, Any]:
        return self.get_init_dict()

    @property
    def mutation_methods(self) -> List[str]:
        return self.layer_mutation_methods + self.node_mutation_methods

    @property
    def layer_mutation_methods(self) -> List[str]:
        return _get_filtered_methods(self, MutationType.LAYER)

    @property
    def node_mutation_methods(self) -> List[str]:
        return _get_filtered_methods(self, MutationType.NODE)

    @property
    def activation(self) -> Optional[str]:
        return None

    @property
    def last_mutation(self) -> Optional[MutationMethod]:
        return self._last_mutation

    @last_mutation.setter
    def last_mutation(self, value: MutationMethod) -> None:
        self._last_mutation = value

    @property
    def last_mutation_attr(self) -> Optional[str]:
        return self._last_mutation_attr

    @last_mutation_attr.setter
    def last_mutation_attr(self, value: str) -> None:
        self._last_mutation_attr = value

    def recreate_network(self, **kwargs) -> None:
        """Recreate the network after a mutation has been applied."""
        if any("." not in method for method in self.mutation_methods):
            raise NotImplementedError(
                "An EvolvableModule must implement the recreate_network method whenever it includes "
                "unique mutation methods."
            )

    def forward(self, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError(
            "forward method must be implemented in order to use the evolvable module."
        )

    def __call__(self, *args, **kwargs) -> torch.Tensor:
        """Forward pass of the network."""
        return self.forward(*args, **kwargs)

    def get_init_dict(self) -> Dict[str, Any]:
        """Get the dictionary of constructor arguments for the network.

        :return: The dictionary of constructor arguments.
        :rtype: Dict[str, Any]
        """
        constructor_args = inspect.signature(self.__init__).parameters

        try:
            return {k: getattr(self, k) for k in constructor_args.keys()}
        except AttributeError:
            raise AttributeError(
                "Custom EvolvableModule objects must be explicit about their "
                "constructor arguments (i.e. don't use *args and **kwargs)"
            )

    def change_activation(self, activation: str, output: bool) -> None:
        """Set the activation function for the network.

        :param activation: Activation function to use.
        :type activation: str
        """
        raise NotImplementedError(
            "change_activation method must be implemented in order to set the activation function."
        )

    def __setattr__(self, name: str, value: Union[Any, SelfEvolvableModule]) -> None:
        """Set attribute of the network. If the attribute is a module, add its mutation methods
        to the network. Otherwise, set the attribute as usual.

        :param name: The name of the attribute.
        :type name: str
        :param value: The value of the attribute.
        :type value: Any
        """
        # Add mutation methods to the network
        if isinstance(value, EvolvableModule):
            if name in self.__dict__["_modules"]:
                self.filter_mutation_methods(name)

            layer_fns = []
            node_fns = []
            for mut_name, method in value.get_mutation_methods().items():
                method_name = ".".join([name, mut_name])
                method_type = method._mutation_type
                if method_type == MutationType.LAYER:
                    layer_fns.append(method_name)
                elif method_type == MutationType.NODE:
                    node_fns.append(method_name)
                else:
                    raise ValueError(f"Invalid mutation type: {method_type}")

            self._layer_mutation_methods += layer_fns
            self._node_mutation_methods += node_fns

        super().__setattr__(name, value)

    def __getattr__(self, name: str) -> Any:
        """Get attribute of the network. If the attribute is a mutation method, return the
        method (from an underlying EvolvableModule attribute also). Otherwise, raise an
        AttributeError.

        :param name: The name of the attribute.
        :type name: str
        :return: The attribute of the network.
        :rtype: Any
        """
        try:
            attr = super().__getattr__(name)
            return attr
        except AttributeError as e:
            mut_method = self.get_mutation_methods().get(name)
            if mut_method is not None:
                return mut_method
            raise e

    def get_output_dense(self) -> Optional[nn.Module]:
        """Get the output dense layer of the network.

        :return: The output dense layer.
        :rtype: nn.Module
        """
        return

    @staticmethod
    def preserve_parameters(old_net: nn.Module, new_net: nn.Module) -> nn.Module:
        """Returns new neural network with copied parameters from old network. Specifically, it
        handles tensors with different sizes by copying the minimum number of elements.

        :param old_net: Old neural network
        :type old_net: nn.Module
        :param new_net: New neural network
        :type new_net: nn.Module
        :return: New neural network with copied parameters
        :rtype: nn.Module
        """
        old_net_dict = dict(old_net.named_parameters())

        for key, param in new_net.named_parameters():
            if key in old_net_dict.keys():
                old_param = old_net_dict[key]
                old_size = old_param.data.size()
                new_size = param.data.size()

                if old_size == new_size:
                    # If the sizes are the same, just copy the parameter
                    param.data = old_param.data
                elif "norm" not in key:
                    # Create a slicing index to handle tensors with varying sizes
                    slice_index = tuple(
                        slice(0, min(o, n)) for o, n in zip(old_size, new_size)
                    )
                    param.data[slice_index] = old_param.data[slice_index]

        return new_net

    @staticmethod
    def init_weights_gaussian(module: nn.Module, std_coeff: float) -> None:
        """Initialize the weights of the neural network using a Gaussian distribution.

        :param module: The neural network module.
        :type module: nn.Module
        :param std_coeff: The standard deviation coefficient.
        :type std_coeff: float
        """

        def init_weights(m):
            if isinstance(m, nn.Linear):
                hidden_size = m.weight.size(1)
                nn.init.normal_(m.weight, mean=0, std=std_coeff / hidden_size)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

        if isinstance(module, nn.Linear):
            init_weights(module)
            return

        layers = [m for m in module.children()]
        for layer in layers:
            init_weights(layer)

    def _init_surface_methods(self) -> None:
        """Initialize the surface mutation methods of the network."""

        def _fetch_methods(cls: Type) -> List[str]:
            return [
                method
                for method in vars(cls).values()
                if isinstance(method, MutationMethod)
            ]

        # Check mutation methods in class
        class_methods: List[MutationMethod] = _fetch_methods(self.__class__)
        layer_methods = []
        node_methods = []
        for method in class_methods:
            if method._mutation_type == MutationType.LAYER:
                layer_methods.append(method.__name__)
            elif method._mutation_type == MutationType.NODE:
                node_methods.append(method.__name__)

        # Check mutation methods in superclasses
        def check_base_methods(cls) -> None:
            for base in cls.__bases__:
                if base is EvolvableModule:
                    return

                base_methods: List[MutationMethod] = _fetch_methods(base)
                for method in base_methods:
                    if method._mutation_type == MutationType.LAYER:
                        layer_methods.append(method.__name__)
                    elif method._mutation_type == MutationType.NODE:
                        node_methods.append(method.__name__)

                check_base_methods(base)

        check_base_methods(self.__class__)

        # We want the unique set of mutation methods across the class and its superclasses
        self._layer_mutation_methods = list(set(layer_methods))
        self._node_mutation_methods = list(set(node_methods))

    def reset_noise(self) -> None:
        """Reset noise for all NoisyLinear layers in the network.

        :param networks: The networks to reset noise for.
        :type networks: nn.Module
        """
        for layer in super().modules():
            if isinstance(layer, NoisyLinear):
                layer.reset_noise()

    def register_mutation_hook(self, hook: Callable) -> None:
        """Register a hook to be called after a mutation has been applied to an
        underlying evolvable module. The hook function should not take any arguments.

        :param hook: The hook function.
        :type hook: Callable
        """
        self._mutation_hook = hook

    def disable_mutations(self, mut_type: Optional[MutationType] = None) -> None:
        """Disable all or some mutation methods from the evolvable module. It recursively
        disables the mutation methods of underlying evolvable modules as well.

        :param mut_type: The type of mutation method to disable.
        :type mut_type: Optional[MutationType]
        """
        if mut_type is None:
            self._layer_mutation_methods = []
            self._node_mutation_methods = []
        elif mut_type == MutationType.LAYER:
            self._layer_mutation_methods = []
        elif mut_type == MutationType.NODE:
            self._node_mutation_methods = []
        else:
            raise ValueError(f"Invalid mutation type: {mut_type}")

        # Recursively disable mutations in underlying evolvable modules
        for module in self.modules().values():
            module.disable_mutations(mut_type)

    def modules(self) -> Dict[str, "EvolvableModule"]:
        """Returns the attributes related to the evolvable modules in the algorithm.
        Includes attributes that are either evolvable modules or a list of evolvable
        modules, as well as the optimizers associated with the networks.

        :return: A dictionary of network attributes.
        :rtype: dict[str, Any]
        """
        # Inspect evolvable
        evolvable_attrs = {}
        for attr in dir(self):
            obj = getattr(self, attr)
            if is_evolvable(attr, obj):
                if isinstance(obj, ModuleDict):
                    for key, mod in obj.items():
                        evolvable_attrs[key] = mod
                else:
                    evolvable_attrs[attr] = obj

        return evolvable_attrs

    def get_mutation_methods(self) -> Dict[str, MutationMethod]:
        """Get all mutation methods for the network.

        :return: A dictionary of mutation methods.
        :rtype: Dict[str, MutationMethod]
        """

        def get_method_from_name(name: str) -> MutationMethod:
            if "." not in name:
                return getattr(self, name)

            attr = name.split(".")[0]
            method = ".".join(name.split(".")[1:])

            return getattr(getattr(self, attr), method)

        return {name: get_method_from_name(name) for name in self.mutation_methods}

    def filter_mutation_methods(self, remove: str) -> None:
        """Filter out mutation methods that contain the specified string in their name.

        param remove: The string to remove.
        type remove: str
        """

        def filter_methods(methods: List[str]) -> List[str]:
            return [method for method in methods if remove not in method]

        self._layer_mutation_methods = filter_methods(self._layer_mutation_methods)
        self._node_mutation_methods = filter_methods(self._node_mutation_methods)

    def get_mutation_probs(self, new_layer_prob: float) -> List[float]:
        """Get the mutation probabilities for each mutation method.

        param new_layer_prob: The probability of selecting a layer mutation method.
        type new_layer_prob: float
        return: A list of probabilities for each mutation method.
        rtype: List[float]
        """
        num_layer_fns = len(self.layer_mutation_methods)
        num_node_fns = len(self.node_mutation_methods)

        num_total = num_layer_fns + num_node_fns
        if num_layer_fns == 0 or num_node_fns == 0:
            return [1 / num_total for _ in range(num_total)]

        probs = []
        for fn in self.get_mutation_methods().values():
            if fn._mutation_type == MutationType.LAYER:
                prob = new_layer_prob / num_layer_fns
            elif fn._mutation_type == MutationType.NODE:
                prob = (1 - new_layer_prob) / num_node_fns

            probs.append(prob)

        return probs

    def sample_mutation_method(
        self, new_layer_prob: float, rng: Optional[Generator] = None
    ) -> MutationMethod:
        """Sample a mutation method based on the mutation probabilities.

        param new_layer_prob: The probability of selecting a layer mutation method.
        type new_layer_prob: float
        param rng: The random number generator.
        type rng: Optional[Generator]
        return: The sampled mutation method.
        rtype: MutationMethod
        """
        if not self.mutation_methods:
            raise ValueError(
                "No mutation methods available. Please use the @mutation decorator to register methods."
            )

        if rng is None:
            rng = np.random.default_rng()

        probs = self.get_mutation_probs(new_layer_prob)
        return rng.choice(self.mutation_methods, p=probs, size=1)[0]

    def clone(self: SelfEvolvableModule) -> SelfEvolvableModule:
        """Returns clone of neural net with identical parameters."""
        clone = self.__class__(**copy.deepcopy(self.get_init_dict()))
        clone._layer_mutation_methods = self._layer_mutation_methods
        clone._node_mutation_methods = self._node_mutation_methods

        # Load state dict if the network has been trained
        try:
            clone.load_state_dict(self.state_dict())
        except RuntimeError:
            pass

        return clone


class EvolvableWrapper(EvolvableModule):
    """Wrapper class for evolvable neural networks. Can be used to provide some
    additional functionality to an EvolvableModule while maintaining its mutation methods
    at the top-level.

    :param module: The evolvable module.
    :type module: EvolvableModule
    """

    def __init__(self, module: EvolvableModule) -> None:
        super().__init__(module.device)

        self._init_wrapped_methods(module, MutationType.LAYER)
        self._init_wrapped_methods(module, MutationType.NODE)

        # Disable mutations in the wrapped module since these are
        # now handled by the wrapper
        module.disable_mutations()
        self._wrapped = module

    @property
    def wrapped(self) -> EvolvableModule:
        return self._wrapped

    def _init_wrapped_methods(
        self, module: EvolvableModule, mut_type: MutationType
    ) -> None:
        """Initialize the mutation methods of the wrapped module.

        :param module: The wrapped module.
        :type module: EvolvableModule
        :param mut_type: The type of mutation method.
        :type mut_type: MutationType
        """
        _fetch = (
            "layer_mutation_methods"
            if mut_type == MutationType.LAYER
            else "node_mutation_methods"
        )

        for method in getattr(module, _fetch):
            if method not in self.mutation_methods:
                setattr(self, method, getattr(module, method))
                current_methods: List[str] = getattr(self, f"_{_fetch}")
                current_methods.append(method)
                setattr(self, f"_{_fetch}", current_methods)
            else:
                raise AttributeError(f"Duplicate mutation method: {method}")

    def modules(self) -> Dict[str, "EvolvableModule"]:
        """Returns the attributes related to the evolvable modules in the algorithm. Includes
        attributes that are either evolvable modules or a list of evolvable modules, as well
        as the optimizers associated with the networks.

        :return: A dictionary of network attributes.
        :rtype: dict[str, Any]
        """
        # Inspect evolvable
        evolvable_attrs = {}
        for attr in dir(self):
            obj = getattr(self, attr)
            if is_evolvable(attr, obj) and attr != "wrapped":
                evolvable_attrs[attr] = obj

        return evolvable_attrs


class ModuleDict(EvolvableModule, nn.ModuleDict):
    """Analogous to ``nn.ModuleDict``, but allows for the recursive inheritance of the
    mutation methods of underlying evolvable modules."""

    @property
    def layer_mutation_methods(self) -> List[str]:
        return [
            f"{name}.{method}"
            for name, module in self.items()
            for method in module.layer_mutation_methods
        ]

    @property
    def node_mutation_methods(self) -> List[str]:
        return [
            f"{name}.{method}"
            for name, module in self.items()
            for method in module.node_mutation_methods
        ]

    def __getitem__(self, key: str) -> EvolvableModule:
        return super().__getitem__(key)

    def values(self) -> Iterable[EvolvableModule]:
        return super().values()

    def items(self) -> Iterable[Tuple[str, EvolvableModule]]:
        return super().items()

    def get_mutation_methods(self) -> Dict[str, MutationMethod]:
        """Get all mutation methods for the network.

        :return: A dictionary of mutation methods.
        :rtype: Dict[str, MutationMethod]
        """

        def get_method_from_name(name: str) -> MutationMethod:
            if "." not in name:
                return getattr(self, name)

            key = name.split(".")[0]
            method = ".".join(name.split(".")[1:])
            return getattr(self[key], method)

        return {name: get_method_from_name(name) for name in self.mutation_methods}



from typing import Any, Dict, Optional

import torch
import torch.distributed as dist
import wandb
from tqdm import trange

from agilerl.algorithms import GRPO
from agilerl.utils.llm_utils import HuggingFaceGym
from agilerl.utils.utils import init_wandb

InitDictType = Optional[Dict[str, Any]]


def finetune_llm(
    agent: GRPO,
    env: HuggingFaceGym,
    checkpoint_interval: Optional[int] = None,
    checkpoint_path: Optional[str] = None,
    wb: bool = False,
    wandb_api_key: Optional[str] = None,
    evaluation_interval: Optional[int] = 10,
    max_reward: Optional[int] = None,
) -> None:
    if wb and agent.local_rank == "0":
        init_wandb(
            algo=agent.algo,
            env_name=env.name,
            wandb_api_key=wandb_api_key,
        )

    if agent.local_rank == "0":
        print("\nTraining...")

    bar_format = "{l_bar}{bar:10}| {n:4}/{total_fmt} [{elapsed:>7}<{remaining:>7}, {rate_fmt}{postfix}]"
    max_steps = len(env) // env.data_batch_size
    if agent.local_rank == "0":
        pbar = trange(
            max_steps,
            unit="step",
            bar_format=bar_format,
            ascii=True,
            dynamic_ncols=True,
        )

    # calling env.reset() supplies the first batch of training data
    prompts = env.reset(reset_dataloaders=True)
    for i in range(max_steps):
        completion_ids, action_masks = agent.get_action(prompts)
        # Use the reward function stored in env.step to calculate reward of the each answer from the group
        next_prompts, rewards = env.step(completion_ids)

        experiences = (
            completion_ids,
            action_masks,
            rewards,
        )
        loss, kl = agent.learn(experiences)
        metrics = [loss, kl, rewards]
        if max_reward is not None:
            accuracy = (rewards == max_reward).sum() / len(rewards.squeeze())
            metrics.append(accuracy)
        agg_metrics = [
            aggregate_metrics_across_gpus(agent, metric) for metric in metrics
        ]
        prompts = next_prompts
        if agent.local_rank == "0":
            metrics = {
                "Loss": (agg_metrics[0]),
                "KL-divergence": (agg_metrics[1]),
                "Mean training reward": (agg_metrics[2]),
            }
            if max_reward is not None:
                metrics |= {"Accuracy": (agg_metrics[3])}
            print(metrics)
            pbar.update(1)
            if wb:
                wandb.log(metrics)
            if (i + 1) % evaluation_interval == 0:
                test_reward = agent.test(env)
                print(f"Test reward: {test_reward}")
                if wb:
                    wandb.log({"Test reward": test_reward})
            if (
                checkpoint_path is not None
                and checkpoint_interval is not None
                and (i + 1) % checkpoint_interval == 0
            ):
                save_llm_checkpoint(agent, checkpoint_path, i)


def gather_tensor(tensor: torch.Tensor, agent: GRPO) -> torch.Tensor:
    # Convert to tensor if it's a scalar
    if not isinstance(tensor, torch.Tensor):
        tensor = torch.tensor(tensor, device=f"cuda:{agent.local_rank}")

    if tensor.device != agent.device:
        tensor = tensor.to(agent.device)
    # Ensure tensor is on correct device
    tensor = tensor.detach().clone()
    # Create a list to store tensors from all processes
    world_size = dist.get_world_size()
    gathered_tensors = [torch.zeros_like(tensor) for _ in range(world_size)]

    # Gather the tensor from all processes
    dist.all_gather(gathered_tensors, tensor)
    return torch.stack(gathered_tensors)


def aggregate_metrics_across_gpus(agent: GRPO, metrics: torch.Tensor):
    all_metrics = gather_tensor(metrics, agent)
    avg_metrics = all_metrics.mean().item()
    return avg_metrics


def save_llm_checkpoint(agent, checkpoint_path, step):
    checkpoint_path = f"step_{step}" if checkpoint_path is None else checkpoint_path
    if agent.accelerator is not None:
        unwrapped_model = agent.accelerator.unwrap_model(agent.actor)
        unwrapped_model.save_pretrained(checkpoint_path)
    else:
        agent.actor.save_pretrained(checkpoint_path)
